# =============================================================================
# Minet FileSystem Utilities
# =============================================================================
#
# Multiple helper functions related to reading and writing files.
#
from typing import Union, Optional, cast, Dict

import os
import gzip
import json
import yaml
from ebbe.decorators import with_defer
from os import makedirs, PathLike
from os.path import basename, join, splitext, abspath, normpath, dirname
from ural import (
    get_hostname,
    get_normalized_hostname,
    get_fingerprinted_hostname,
    safe_urlsplit,
    pathsplit,
)
from quenouille import NamedLocks

from minet.exceptions import (
    FilenameFormattingError,
    DefinitionInvalidFormatError,
    CouldNotInferEncodingError,
)
from minet.utils import md5, PseudoFStringFormatter
from minet.encodings import infer_encoding
from minet.sqlar import SQLiteArchive


def read_potentially_gzipped_path(
    path,
    encoding: Optional[str] = None,
    errors: str = "replace",
    fallback_encoding: Optional[str] = None,
) -> str:
    open_fn = open
    flag = "r" if encoding is not None else "rb"

    if path.endswith(".gz"):
        open_fn = gzip.open
        flag = "rt" if encoding is not None else "r"

    if encoding is None:
        with open_fn(path, flag) as f:
            binary = f.read()
            encoding = infer_encoding(binary)

            if encoding is None:
                if fallback_encoding is not None:
                    encoding = fallback_encoding
                else:
                    raise CouldNotInferEncodingError

            return binary.decode(encoding, errors=errors)

    with open_fn(path, flag, encoding=encoding, errors=errors) as f:
        return f.read()


@with_defer()
def load_definition(f, *, defer=None, encoding: str = "utf-8") -> Dict:
    if isinstance(f, (str, PathLike)):
        path = str(f)
        f = open(path, encoding=encoding)
        defer(f.close)  # type: ignore
    else:
        path = f.name

    if path.endswith(".json"):
        definition = json.load(f)

    elif path.endswith(".yml") or path.endswith(".yaml"):
        definition = yaml.safe_load(f)

    else:
        raise DefinitionInvalidFormatError

    return definition


class FolderStrategy(object):
    CHOICES = ["flat", "fullpath", "hostname", "normalize-hostname", "prefix-x"]

    DOCUMENTATION = """
        . "flat": default choice, all files will be written in the indicated
            content folder.

        . "fullpath": all files will be written in a folder consisting of the
            url hostname and then its path.

        . "prefix-x": e.g. "prefix-4", files will be written in folders
            having a name that is the first x characters of the file's name.
            This is an efficient way to partition content into folders containing
            roughly the same number of files if the file names are random (which
            is the case by default since md5 hashes will be used).

        . "hostname": files will be written in folders based on their url's
            full host name.

        . "normalized-hostname": files will be written in folders based on
            their url's hostname stripped of some undesirable parts (such as
            "www.", or "m.", for instance).

        . "fingerprinted-hostname": files will be written in folders based on
            their url's hostname stripped of some more undesirable parts (such as
            "fr.", for instance) and their public suffix will be dropped.
    """

    def __call__(self):
        raise NotImplementedError

    @staticmethod
    def from_name(name):
        if name == "flat":
            return FlatFolderStrategy()

        if name == "fullpath":
            return FullPathFolderStrategy()

        if name == "hostname":
            return HostnameFolderStrategy()

        if name == "normalized-hostname":
            return NormalizedHostnameFolderStrategy()

        if name == "fingerprinted-hostname":
            return FingerprintedHostnameFolderStrategy()

        if name.startswith("prefix-"):
            length = name.split("prefix-")[-1]

            try:
                length = int(length)
            except ValueError:
                raise TypeError

            if length <= 0:
                raise TypeError

            return PrefixFolderStrategy(length)

        raise TypeError


class FlatFolderStrategy(FolderStrategy):
    def __call__(self, filename, **kwargs):
        return filename


class FullPathFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        parsed = safe_urlsplit(url)
        final_path = [cast(str, parsed.hostname)]

        path = pathsplit(parsed.path)

        if path and "." in path[-1]:
            path.pop()

        final_path += path

        # NOTE: dealing with typical max filename length
        for i in range(len(final_path)):
            item = final_path[i]

            if len(item) > 255:
                final_path[i] = item[:255]

        return join(os.sep.join(final_path), filename)


class PrefixFolderStrategy(FolderStrategy):
    def __init__(self, length):
        self.length = length

    def __call__(self, filename, **kwargs):
        base = basename(filename).split(".", 1)[0]

        return join(base[: self.length], filename)


class HostnameFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        hostname = get_hostname(url)

        if not hostname:
            hostname = "unknown-host"

        return join(hostname, filename)


class NormalizedHostnameFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        hostname = get_normalized_hostname(
            url,
            normalize_amp=False,
            infer_redirection=False,
        )

        if not hostname:
            hostname = "unknown-host"

        return join(hostname, filename)


class FingerprintedHostnameFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        hostname = get_fingerprinted_hostname(url, strip_suffix=True)

        if not hostname:
            hostname = "unknown-host"

        return join(hostname, filename)


class FilenameBuilder(object):
    folder_strategy: Optional[FolderStrategy]

    def __init__(
        self,
        folder_strategy: Optional[Union[str, FolderStrategy]] = None,
        template=None,
    ):
        self.folder_strategy = None

        if folder_strategy is not None and isinstance(folder_strategy, str):
            self.folder_strategy = FolderStrategy.from_name(folder_strategy)
        elif isinstance(folder_strategy, FolderStrategy):
            self.folder_strategy = folder_strategy

        self.formatter = PseudoFStringFormatter()
        self.template = template

    def __call__(
        self, url=None, filename=None, ext=None, formatter_kwargs={}, compressed=False
    ):
        original_ext = None

        if filename is None:
            base = md5(url)
        else:
            base, original_ext = splitext(filename)

        # We favor the extension found in given filename, else we fallback
        # on the provided one if any (usually inferred from http response)
        ext = original_ext if original_ext else (ext or "")

        # Extension normalization
        if ext:
            ext = "." + ext.lstrip(".")

        if self.template is not None:
            try:
                filename = self.formatter.format(
                    self.template, value=base, ext=ext, **formatter_kwargs
                )
            except Exception as e:
                raise FilenameFormattingError(reason=e, template=self.template)
        else:
            filename = base + ext

        if self.folder_strategy:
            filename = self.folder_strategy(filename, url=url)

        if compressed:
            filename += ".gz"

        return filename


class ThreadSafeFileWriter(object):
    def __init__(self, root_directory: Optional[str] = None, sqlar: bool = False):
        self.root_directory = root_directory or ""
        self.folder_locks = NamedLocks()
        self.file_locks = NamedLocks()
        self.sqlar = sqlar
        self.archive = None

        if self.sqlar:
            self.root_directory += ".sqlar"
            self.archive = SQLiteArchive(self.root_directory)

    def resolve(self, filename: str, relative: bool = False, compress: bool = False):
        full_path = join(self.root_directory, filename)

        if compress and not full_path.endswith(".gz"):
            full_path += ".gz"

        if relative:
            return normpath(full_path)

        return normpath(abspath(full_path))

    def makedirs(self, directory: str) -> None:
        if not directory:
            return

        # TODO: cache
        with self.folder_locks[directory]:
            makedirs(directory, exist_ok=True)

    def write(
        self,
        filename: str,
        contents: Union[str, bytes],
        compress: bool = False,
        relative: bool = False,
    ) -> str:
        if self.sqlar:
            assert self.archive is not None

            if not isinstance(contents, bytes):
                contents = contents.encode("utf-8")

            self.archive.write(filename, contents)

            return filename

        binary = isinstance(contents, bytes)
        filename = self.resolve(filename, relative=relative, compress=compress)
        directory = dirname(filename)

        # NOTE: Could have prefix-free locking as a bonus...
        self.makedirs(directory)

        open_kwargs = {"mode": "wb" if binary else "w"}

        if not binary:
            open_kwargs["encoding"] = "utf-8"

        if compress:
            open_fn = gzip.open

            if not binary:
                open_kwargs["mode"] = "wt"
        else:
            open_fn = open

        with self.file_locks[filename]:
            with open_fn(filename, **open_kwargs) as f:
                f.write(contents)  # type: ignore

        return filename



from minet.exceptions import MinetError


class WikipediaError(MinetError):
    pass


class WikimediaRESTAPIError(WikipediaError):
    pass


class WikimediaRESTAPIThrottledError(WikimediaRESTAPIError):
    pass


class WikimediaRESTAPIServerError(WikimediaRESTAPIError):
    pass



#!/usr/bin/env python
# =============================================================================
# Minet CLI Endpoint
# =============================================================================
#
# CLI endpoint of the Minet library.
#
from typing import Optional, List

import csv
import sys
import ctypes
import importlib
import logging
import multiprocessing
import casanova
import shlex
import platform
from casanova.exceptions import MissingColumnError
from contextlib import ExitStack

# NOTE: this is necessary for pyinstaller build still...
from encodings import idna  # noqa: F401

from minet.cli.constants import DEFAULT_PREBUFFER_BYTES
from minet.loggers import sleepers_logger, downloaders_logger
from minet.cli.console import console
from minet.cli.utils import (
    die,
    get_rcfile,
    CLIRetryerHandler,
    CLIDownloaderHandler,
    with_cli_exceptions,
)
from minet.cli.argparse import resolve_arg_dependencies, build_parser, get_subparser
from minet.cli.exceptions import NotResumableError, InvalidArgumentsError, FatalError

WINDOWS = "windows" in platform.system().lower()

GLOBAL_SETUP_IS_DONE = False


def global_setup() -> None:
    global GLOBAL_SETUP_IS_DONE

    if GLOBAL_SETUP_IS_DONE:
        return

    GLOBAL_SETUP_IS_DONE = True

    # Issue #497, utf-8 encoding for windows stdout
    if WINDOWS:
        sys.__stdout__.reconfigure(encoding="utf-8")
        sys.__stderr__.reconfigure(encoding="utf-8")

    # Freezing multiprocessing support for pyinstaller etc.
    multiprocessing.freeze_support()

    # Default spawn context for multiprocessing
    multiprocessing.set_start_method("spawn")

    # Increasing max CSV file limit to avoid pesky issues
    csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))

    # Casanova global defaults
    casanova.set_defaults(
        prebuffer_bytes=DEFAULT_PREBUFFER_BYTES,
        strip_null_bytes_on_read=True,
        strip_null_bytes_on_write=True,
    )

    # Adding handlers for various loggers
    sleepers_logger.addHandler(CLIRetryerHandler())
    downloaders_logger.addHandler(CLIDownloaderHandler())
    downloaders_logger.setLevel(level=logging.INFO)


@with_cli_exceptions
def run(name: str, version: str, commands: List, args: Optional[str] = None):
    global_setup()

    # Building parser
    parser, subparser_index = build_parser(name, version, commands)

    # Parsing arguments and triggering commands
    cli_args = parser.parse_args(shlex.split(args) if args is not None else None)

    # Suppressing console?
    if getattr(cli_args, "silent", False):
        console.quiet = True

    action = subparser_index.get(cli_args.action)

    if action is not None:
        # If subparser is called without any subaction, we print help and exit
        if "subparsers" in action["command"]:
            subdest = action["command"]["subparsers"]["dest"]
            subaction = getattr(cli_args, subdest)

            if subaction is None:
                action["parser"].print_help()
                sys.exit(0)

        # Loading config
        config = get_rcfile(cli_args.rcfile) if hasattr(cli_args, "rcfile") else None

        # Resolving namespace dependencies
        try:
            to_close = resolve_arg_dependencies(cli_args, config)
        except OSError as e:
            parser.error("Could not open output file (-o/--output): %s" % str(e))
        except InvalidArgumentsError as e:
            parser.error(e.message)
        except NotResumableError:
            parser.error(
                "Cannot --resume without knowing where the output will be written (use -o/--output)"
            )

        # Lazy loading module for faster startup
        meta = action["command"]

        if hasattr(cli_args, "subcommand") and cli_args.subcommand:
            meta = action["command"]["subparsers"]["commands"][cli_args.subcommand]

        if "resolve" in meta:
            try:
                meta["resolve"](cli_args)
            except InvalidArgumentsError as e:
                parser.error(e.message)

        fn = meta["package"]

        # Lazy-loading
        if not callable(fn):
            m = importlib.import_module(meta["package"])
            fn = getattr(m, "action")

        with ExitStack() as stack:
            stack.callback(sys.stdout.flush)
            stack.callback(sys.stderr.flush)

            for buffer in to_close:
                stack.callback(buffer.close)

            try:
                fn(cli_args)
            except InvalidArgumentsError as e:
                parser.error(e.message)
            except MissingColumnError as e:
                console.print("Missing column", "[error]{}[/error]!".format(str(e)))
                sys.exit(1)
            except FatalError as e:
                console.vprint(e.message, style="error" if not e.warning else "warning")
                sys.exit(1)

    elif cli_args.action == "help":
        if len(cli_args.subcommand) == 0:
            parser.print_help()
            return

        target = get_subparser(subparser_index, cli_args.subcommand)

        if target is None:
            die('Unknow command "%s"' % " ".join(cli_args.subcommand))
        else:
            target.print_help()

    else:
        parser.print_help()



from typing import List

import re
from playwright.async_api import BrowserContext, Response, TimeoutError, Locator
from playwright_stealth import stealth_async
from ural.facebook import has_facebook_comments

from minet.browser import ThreadsafeBrowser
from minet.browser.utils import try_expect_response

from minet.facebook.exceptions import FacebookInvalidTargetError
from minet.facebook.types import FacebookComment


VIEW_MORE_COMMENTS_RE = re.compile(r"View\s+.+more\s+comments?", re.I)
VIEW_MORE_REPLIES_RE = re.compile(r"View\s+.+more\s+replies", re.I)
VIEW_MORE_SUBREPLIES_RE = re.compile(r"^\s*\d+\s+replies", re.I)


def is_graphql_comments_response(response: Response) -> bool:
    if "/api/graphql/" not in response.url:
        return False

    friendly_name = response.request.headers.get("x-fb-friendly-name")

    return (
        friendly_name == "CometUFICommentsProviderForDisplayCommentsQuery"
        or friendly_name == "CometUFIFullThreadedSubRepliesListDataProviderQuery"
    )


async def try_to_click(locator: Locator, timeout: int = 500) -> None:
    try:
        await locator.click(timeout=timeout)
    except TimeoutError:
        pass


class FacebookEmulatedScraper:
    def __init__(self, headless=True):
        self.browser = ThreadsafeBrowser(headless=headless, width=1024, height=768)

    def __enter__(self):
        return self

    def stop(self) -> None:
        self.browser.stop()

    def __exit__(self, *args):
        self.stop()

    async def __scrape_comments(
        self, context: BrowserContext, url: str
    ) -> List[FacebookComment]:
        async with await context.new_page() as page:
            await stealth_async(page)

            await page.goto(url)

            await try_to_click(page.get_by_label("Decline optional cookies").first)
            await try_to_click(page.get_by_label("Close").first)

            # NOTE: this removes the overlay prompting you to login. This is
            # useful for debug, but not only because we need to be able to
            # click & scroll the page and the overlay prevents us from doing so.
            await page.evaluate(
                """
                () => {
                    document.querySelector('[data-nosnippet]').remove();
                }
                """
            )

            async def expect_comments(action) -> List[FacebookComment]:
                response = await try_expect_response(
                    page, action, is_graphql_comments_response
                )

                payload = await response.json()

                # with open("./dump.json", "w") as f:
                #     import json

                #     json.dump(payload, f, ensure_ascii=False, indent=2)

                return FacebookComment.from_payload(payload)

            async def select_all_comments():
                await page.get_by_text("Most relevant").first.click()
                await page.get_by_text("All comments").first.click()

            async def view_more_comments():
                await page.get_by_text(VIEW_MORE_COMMENTS_RE).first.click(timeout=3000)

            async def view_more_replies():
                await page.get_by_text(VIEW_MORE_REPLIES_RE).first.click(timeout=3000)

            async def view_more_subreplies():
                await page.get_by_text(VIEW_MORE_SUBREPLIES_RE).first.click(
                    timeout=3000
                )

            comments: List[FacebookComment] = []

            comments.extend(await expect_comments(select_all_comments))

            # Deploying comments
            while True:
                await page.wait_for_timeout(1000)

                try:
                    comments.extend(await expect_comments(view_more_comments))
                except TimeoutError:
                    break

            # Deploying replies
            while True:
                await page.wait_for_timeout(1000)

                try:
                    comments.extend(await expect_comments(view_more_replies))
                except TimeoutError:
                    break

            # Deploying subreplies
            while True:
                await page.wait_for_timeout(1000)

                try:
                    comments.extend(await expect_comments(view_more_subreplies))
                except TimeoutError:
                    break

            # await page.wait_for_timeout(1000000)

            return FacebookComment.sort(comments)

    def scrape_comments(self, url: str) -> List[FacebookComment]:
        if not has_facebook_comments(url):
            raise FacebookInvalidTargetError

        return self.browser.run_in_default_context(self.__scrape_comments, url)



# =============================================================================
# Minet HTTP Multithreaded Executors
# =============================================================================
#
# Exposing a specialized quenouille wrapper grabbing various urls from the
# web in a multithreaded fashion.
#
from typing import (
    cast,
    overload,
    Optional,
    Iterator,
    Callable,
    TypeVar,
    Generic,
    Iterable,
    Dict,
    Union,
    Tuple,
    Awaitable,
    Any,
    TYPE_CHECKING,
)
from minet.types import Literal, TypedDict, Unpack, NotRequired

if TYPE_CHECKING:
    from minet.browser.threadsafe_browser import BrowserOrBrowserContext

import urllib3
import threading
from threading import Event
from quenouille import ThreadPoolExecutor
from ural import get_domain_name, ensure_protocol
from tenacity import RetryCallState

from minet.serialization import serialize_error_as_slug
from minet.exceptions import CancelledRequestError, HTTPCallbackError
from minet.web import (
    create_pool_manager,
    create_request_retryer,
    request,
    resolve,
    Response,
    RedirectionStack,
    AnyTimeout,
    EXPECTED_WEB_ERRORS,
)
from minet.constants import (
    DEFAULT_DOMAIN_PARALLELISM,
    DEFAULT_IMAP_BUFFER_SIZE,
    DEFAULT_THROTTLE,
    DEFAULT_URLLIB3_TIMEOUT,
    DEFAULT_FETCH_MAX_REDIRECTS,
    DEFAULT_RESOLVE_MAX_REDIRECTS,
)

ItemType = TypeVar("ItemType")
ResultType = TypeVar("ResultType")
CallbackResultType = TypeVar("CallbackResultType")


class HTTPWorkerPayloadBase(Generic[ItemType]):
    __slots__ = ("item", "url", "__has_cached_domain", "__domain")

    item: ItemType
    url: Optional[str]

    __has_cached_domain: bool
    __domain: Optional[str]

    def __init__(self, item: ItemType, url: Optional[str]):
        self.item = item
        self.url = url
        self.__has_cached_domain = False
        self.__domain = None

    @property
    def domain(self) -> Optional[str]:
        if self.__has_cached_domain:
            return self.__domain

        if self.url is not None:
            self.__domain = get_domain_name(self.url)

        self.__has_cached_domain = True

        return self.__domain


class HTTPWorkerPayload(HTTPWorkerPayloadBase[ItemType]):
    item: ItemType
    url: str


ArgsCallbackType = Callable[[HTTPWorkerPayload[ItemType]], Dict]


class ExecutorRequestKwargs(TypedDict, Generic[ItemType, CallbackResultType]):
    ordered: NotRequired[bool]
    key: NotRequired[Optional[Callable[[ItemType], Optional[str]]]]
    throttle: NotRequired[float]
    request_args: NotRequired[Optional[ArgsCallbackType[ItemType]]]
    use_pycurl: NotRequired[bool]
    compressed: NotRequired[bool]
    buffer_size: NotRequired[int]
    domain_parallelism: NotRequired[int]
    max_redirects: NotRequired[int]


class ExecutorResolveKwargs(TypedDict, Generic[ItemType, CallbackResultType]):
    ordered: NotRequired[bool]
    key: NotRequired[Optional[Callable[[ItemType], Optional[str]]]]
    throttle: NotRequired[float]
    resolve_args: NotRequired[Optional[ArgsCallbackType[ItemType]]]
    buffer_size: NotRequired[int]
    domain_parallelism: NotRequired[int]
    max_redirects: NotRequired[int]
    follow_refresh_header: NotRequired[bool]
    follow_meta_refresh: NotRequired[bool]
    follow_js_relocation: NotRequired[bool]
    infer_redirection: NotRequired[bool]
    canonicalize: NotRequired[bool]


class RequestResult(Generic[ItemType]):
    __slots__ = ("item", "url", "error", "response")

    item: ItemType
    url: Optional[str]
    error: Optional[Exception]
    response: Optional[Response]

    def __init__(self, item: ItemType, url: Optional[str]):
        self.item = item
        self.url = url
        self.error = None
        self.response = None

    @property
    def error_code(self) -> Optional[str]:
        return serialize_error_as_slug(self.error) if self.error is not None else None

    def __repr__(self):
        name = self.__class__.__name__

        if not self.url:
            return "<{name} null!>".format(name=name)

        if not self.response:
            return "<{name} url={url!r} pending!>".format(name=name, url=self.url)

        if self.error:
            return "<{name} url={url!r} error={error}>".format(
                name=name, url=self.url, error=self.error.__class__.__name__
            )

        assert self.response is not None

        return "<{name} url={url!r} status={status!r}>".format(
            name=name, url=self.url, status=self.response.status
        )


class PassthroughRequestResult(RequestResult[ItemType]):
    item: ItemType
    url: None
    error: None
    response: None

    def __init__(self, item: ItemType):
        self.item = item
        self.url = None
        self.error = None
        self.response = None

    @property
    def error_code(self) -> None:
        return None


class ErroredRequestResult(RequestResult[ItemType]):
    item: ItemType
    url: str
    error: Exception
    response: None

    def __init__(self, item: ItemType, url: str, error: Exception):
        self.item = item
        self.url = url
        self.error = error
        self.response = None

    @property
    def error_code(self) -> str:
        return serialize_error_as_slug(self.error)


class SuccessfulRequestResult(RequestResult[ItemType]):
    item: ItemType
    url: str
    error: None
    response: Response

    def __init__(self, item: ItemType, url: str, response: Response):
        self.item = item
        self.url = url
        self.error = None
        self.response = response

    @property
    def error_code(self) -> None:
        return None


AnyActualRequestResult = Union[
    ErroredRequestResult[ItemType], SuccessfulRequestResult[ItemType]
]

AnyRequestResult = Union[
    AnyActualRequestResult[ItemType], PassthroughRequestResult[ItemType]
]


class ResolveResult(Generic[ItemType]):
    __slots__ = ("item", "url", "error", "stack")

    item: ItemType
    url: Optional[str]
    error: Optional[Exception]
    stack: Optional[RedirectionStack]

    def __init__(self, item: ItemType, url: Optional[str]):
        self.item = item
        self.url = url
        self.error = None
        self.stack = None

    @property
    def error_code(self) -> Optional[str]:
        return serialize_error_as_slug(self.error) if self.error is not None else None

    def __repr__(self):
        name = self.__class__.__name__

        if not self.url:
            return "<{name} null!>".format(name=name)

        if self.error:
            return "<{name} url={url!r} error={error}>".format(
                name=name, url=self.url, error=self.error.__class__.__name__
            )

        assert self.stack is not None

        return "<{name} url={url!r} status={status!r} redirects={redirects!r}>".format(
            name=name,
            url=self.url,
            status=self.stack[-1].status,
            redirects=len(self.stack),
        )


class PassthroughResolveResult(ResolveResult[ItemType]):
    item: ItemType
    url: None
    error: None
    stack: None

    def __init__(self, item: ItemType):
        self.item = item
        self.url = None
        self.error = None
        self.stack = None

    @property
    def error_code(self) -> None:
        return None


class ErroredResolveResult(ResolveResult[ItemType]):
    item: ItemType
    url: str
    error: Exception
    stack: None

    def __init__(self, item: ItemType, url: str, error: Exception):
        self.item = item
        self.url = url
        self.error = error
        self.stack = None

    @property
    def error_code(self) -> str:
        return serialize_error_as_slug(self.error)


class SuccessfulResolveResult(ResolveResult[ItemType]):
    item: ItemType
    url: str
    error: None
    stack: RedirectionStack

    def __init__(self, item: ItemType, url: str, stack: RedirectionStack):
        self.item = item
        self.url = url
        self.error = None
        self.stack = stack

    @property
    def error_code(self) -> None:
        return None


AnyActualResolveResult = Union[
    ErroredResolveResult[ItemType], SuccessfulResolveResult[ItemType]
]

AnyResolveResult = Union[
    AnyActualResolveResult[ItemType], PassthroughResolveResult[ItemType]
]


def key_by_domain_name(payload: HTTPWorkerPayloadBase) -> Optional[str]:
    return payload.domain


def payloads_iter(
    iterable: Iterable[ItemType],
    key: Optional[Callable[[ItemType], Optional[str]]] = None,
    passthrough: bool = False,
) -> Iterator[HTTPWorkerPayloadBase[ItemType]]:
    for item in iterable:
        url = item if key is None else key(item)

        if not url:
            if not passthrough:
                raise TypeError("item has no url: {!r}".format(item))

            yield HTTPWorkerPayloadBase(item=item, url=None)
            continue

        # Url cleanup
        url = ensure_protocol(url.strip())  # type: ignore

        yield HTTPWorkerPayloadBase(item=item, url=url)


class HTTPWorker(Generic[ItemType, CallbackResultType]):
    def __init__(
        self,
        pool_manager: urllib3.PoolManager,
        cancel_event: Event,
        local_context: threading.local,
        *,
        resolving: bool = False,
        get_args: Optional[ArgsCallbackType[ItemType]] = None,
        use_pycurl: bool = False,
        compressed: bool = False,
        max_redirects: int = DEFAULT_FETCH_MAX_REDIRECTS,
        follow_refresh_header: bool = True,
        follow_meta_refresh: bool = False,
        follow_js_relocation: bool = False,
        infer_redirection: bool = False,
        canonicalize: bool = False,
        callback: Optional[
            Union[
                Callable[[ItemType, str, Response], CallbackResultType],
                Callable[[ItemType, str, RedirectionStack], CallbackResultType],
            ]
        ] = None,
    ):
        self.cancel_event = cancel_event
        self.local_context = local_context
        self.get_args = get_args
        self.callback = callback

        self.resolving = resolving
        self.fn = request if not resolving else resolve
        self.PassthroughResult = (
            PassthroughRequestResult if not resolving else PassthroughResolveResult
        )
        self.SuccessfulResult = (
            SuccessfulRequestResult if not resolving else SuccessfulResolveResult
        )
        self.ErroredResult = (
            ErroredRequestResult if not resolving else ErroredResolveResult
        )

        self.default_kwargs = {
            "pool_manager": pool_manager,
            "max_redirects": max_redirects,
            "cancel_event": cancel_event,
            "follow_refresh_header": follow_refresh_header,
            "follow_meta_refresh": follow_meta_refresh,
            "follow_js_relocation": follow_js_relocation,
            "infer_redirection": infer_redirection,
            "canonicalize": canonicalize,
        }

        if use_pycurl:
            del self.default_kwargs["pool_manager"]
            self.default_kwargs["use_pycurl"] = True

        if compressed:
            self.default_kwargs["compressed"] = True

    def __call__(
        self, payload: HTTPWorkerPayloadBase[ItemType]
    ) -> Optional[
        Tuple[
            Union[
                AnyRequestResult[ItemType],
                AnyResolveResult[ItemType],
            ],
            Optional[CallbackResultType],
        ]
    ]:
        item, url = payload.item, payload.url

        # Noop
        if url is None:
            return self.PassthroughResult(item), None  # type: ignore

        kwargs = self.default_kwargs.copy()

        if self.cancel_event.is_set():
            return

        if self.get_args is not None:
            # NOTE: given callback must be threadsafe
            kwargs.update(self.get_args(cast(HTTPWorkerPayload, payload)))

        if self.cancel_event.is_set():
            return

        try:
            retryer = getattr(self.local_context, "retryer", None)

            if retryer is not None:
                output = retryer(self.fn, url, **kwargs)
            else:
                output = self.fn(url, **kwargs)

        except CancelledRequestError:
            return

        except EXPECTED_WEB_ERRORS as error:
            return self.ErroredResult(item, url, error), None

        else:
            callback_result = None

            if self.callback is not None:
                if self.cancel_event.is_set():
                    return

                try:
                    if retryer is not None:
                        callback_result = retryer(self.callback, item, url, output)
                    else:
                        callback_result = self.callback(item, url, output)  # type: ignore
                except Exception as reason:
                    return (
                        self.ErroredResult(item, url, HTTPCallbackError(reason)),
                        None,
                    )

        return self.SuccessfulResult(item, url, output), callback_result  # type: ignore


class HTTPThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: Optional[int] = None,
        wait: bool = True,
        daemonic: bool = False,
        insecure: bool = False,
        timeout: AnyTimeout = DEFAULT_URLLIB3_TIMEOUT,
        spoof_tls_ciphers: bool = False,
        proxy: Optional[str] = None,
        retry: bool = False,
        retryer_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.cancel_event = Event()
        self.local_context = threading.local()

        if retry:

            def epilog(retry_state: RetryCallState) -> str:
                return retry_state.args[0]

            default_retryer_kwargs = {
                "retry_on_timeout": False,
                "cancel_event": self.cancel_event,
                "max_attempts": 3,
                "epilog": epilog,
            }

            default_retryer_kwargs.update(retryer_kwargs or {})

            def init_local_context():
                self.local_context.retryer = create_request_retryer(
                    **default_retryer_kwargs
                )

            kwargs["initializer"] = init_local_context

        super().__init__(
            max_workers,
            wait=wait,
            daemonic=daemonic,
            **kwargs,
        )

        # NOTE: 0 workers means a synchronous pool in quenouille,
        # so we reserve at least one connection for the pool.
        self.pool_manager = create_pool_manager(
            parallelism=max(1, self.max_workers),
            num_pools=max(64, self.max_workers),
            insecure=insecure,
            timeout=timeout,
            spoof_tls_ciphers=spoof_tls_ciphers,
            proxy=proxy,
        )

    def cancel(self) -> None:
        self.cancel_event.set()

    def shutdown(self, wait=True) -> None:
        self.cancel()
        self.pool_manager.clear()
        return super().shutdown(wait=wait)

    @overload
    def request(
        self,
        iterator: Iterable[ItemType],
        *,
        passthrough: Literal[False] = ...,
        callback: None = ...,
        **kwargs: Unpack[ExecutorRequestKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[AnyActualRequestResult[ItemType]]:
        ...

    @overload
    def request(
        self,
        iterator: Iterable[ItemType],
        *,
        passthrough: Literal[True] = ...,
        callback: None = ...,
        **kwargs: Unpack[ExecutorRequestKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[AnyRequestResult[ItemType]]:
        ...

    @overload
    def request(
        self,
        iterator: Iterable[ItemType],
        *,
        passthrough: Literal[False] = ...,
        callback: Callable[[ItemType, str, Response], CallbackResultType] = ...,
        **kwargs: Unpack[ExecutorRequestKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[Tuple[AnyActualRequestResult[ItemType], CallbackResultType]]:
        ...

    @overload
    def request(
        self,
        iterator: Iterable[ItemType],
        *,
        passthrough: Literal[True] = ...,
        callback: Callable[[ItemType, str, Response], CallbackResultType] = ...,
        **kwargs: Unpack[ExecutorRequestKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[Tuple[AnyRequestResult[ItemType], CallbackResultType]]:
        ...

    def request(
        self,
        iterator: Iterable[ItemType],
        *,
        ordered: bool = False,
        key: Optional[Callable[[ItemType], Optional[str]]] = None,
        throttle: float = DEFAULT_THROTTLE,
        request_args: Optional[ArgsCallbackType[ItemType]] = None,
        use_pycurl: bool = False,
        compressed: bool = False,
        buffer_size: int = DEFAULT_IMAP_BUFFER_SIZE,
        domain_parallelism: int = DEFAULT_DOMAIN_PARALLELISM,
        max_redirects: int = DEFAULT_FETCH_MAX_REDIRECTS,
        callback: Optional[
            Callable[[ItemType, str, Response], CallbackResultType]
        ] = None,
        passthrough: bool = False,
    ) -> Union[
        Iterator[AnyRequestResult[ItemType]],
        Iterator[AnyActualRequestResult[ItemType]],
        Iterator[Tuple[AnyRequestResult[ItemType], CallbackResultType]],
        Iterator[Tuple[AnyActualRequestResult[ItemType], CallbackResultType]],
    ]:
        # TODO: validate
        worker = HTTPWorker(
            self.pool_manager,
            self.cancel_event,
            self.local_context,
            get_args=request_args,
            max_redirects=max_redirects,
            use_pycurl=use_pycurl,
            compressed=compressed,
            callback=callback,
        )

        method = super().imap if ordered else super().imap_unordered

        imap = method(
            payloads_iter(iterator, key=key, passthrough=passthrough),
            worker,
            key=key_by_domain_name,
            parallelism=domain_parallelism,
            buffer_size=buffer_size,
            throttle=throttle,
        )

        for item in imap:
            if item is None:
                continue

            if callback is not None:
                yield item  # type: ignore
            else:
                yield item[0]  # type: ignore

    @overload
    def resolve(
        self,
        iterator: Iterable[ItemType],
        *,
        callback: None = ...,
        passthrough: Literal[False] = ...,
        **kwargs: Unpack[ExecutorResolveKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[AnyActualResolveResult[ItemType]]:
        ...

    @overload
    def resolve(
        self,
        iterator: Iterable[ItemType],
        *,
        callback: None = ...,
        passthrough: Literal[True] = ...,
        **kwargs: Unpack[ExecutorResolveKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[AnyResolveResult[ItemType]]:
        ...

    @overload
    def resolve(
        self,
        iterator: Iterable[ItemType],
        *,
        callback: Callable[[ItemType, str, RedirectionStack], CallbackResultType],
        passthrough: Literal[False] = ...,
        **kwargs: Unpack[ExecutorResolveKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[Tuple[AnyActualResolveResult[ItemType], CallbackResultType]]:
        ...

    @overload
    def resolve(
        self,
        iterator: Iterable[ItemType],
        *,
        callback: Callable[[ItemType, str, RedirectionStack], CallbackResultType],
        passthrough: Literal[True] = ...,
        **kwargs: Unpack[ExecutorResolveKwargs[ItemType, CallbackResultType]],
    ) -> Iterator[Tuple[AnyResolveResult[ItemType], CallbackResultType]]:
        ...

    def resolve(
        self,
        iterator: Iterable[ItemType],
        *,
        ordered: bool = False,
        key: Optional[Callable[[ItemType], Optional[str]]] = None,
        throttle: float = DEFAULT_THROTTLE,
        resolve_args: Optional[ArgsCallbackType[ItemType]] = None,
        buffer_size: int = DEFAULT_IMAP_BUFFER_SIZE,
        domain_parallelism: int = DEFAULT_DOMAIN_PARALLELISM,
        max_redirects: int = DEFAULT_RESOLVE_MAX_REDIRECTS,
        follow_refresh_header: bool = True,
        follow_meta_refresh: bool = False,
        follow_js_relocation: bool = False,
        infer_redirection: bool = False,
        canonicalize: bool = False,
        callback: Optional[
            Callable[[ItemType, str, RedirectionStack], CallbackResultType]
        ] = None,
        passthrough: bool = False,
    ) -> Union[
        Iterator[AnyResolveResult[ItemType]],
        Iterator[AnyActualResolveResult[ItemType]],
        Iterator[Tuple[AnyResolveResult[ItemType], CallbackResultType]],
        Iterator[Tuple[AnyActualResolveResult[ItemType], CallbackResultType]],
    ]:
        # TODO: validate
        worker = HTTPWorker(
            self.pool_manager,
            self.cancel_event,
            self.local_context,
            resolving=True,
            get_args=resolve_args,
            max_redirects=max_redirects,
            follow_refresh_header=follow_refresh_header,
            follow_meta_refresh=follow_meta_refresh,
            follow_js_relocation=follow_js_relocation,
            infer_redirection=infer_redirection,
            canonicalize=canonicalize,
            callback=callback,
        )

        method = super().imap if ordered else super().imap_unordered

        imap = method(
            payloads_iter(iterator, key=key, passthrough=passthrough),
            worker,
            key=key_by_domain_name,
            parallelism=domain_parallelism,
            buffer_size=buffer_size,
            throttle=throttle,
        )

        for item in imap:
            if item is None:
                continue

            if callback is not None:
                yield item  # type: ignore
            else:
                yield item[0]  # type: ignore


class BrowserThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: Optional[int] = None,
        wait: bool = True,
        daemonic: bool = False,
        height: int = 1920,
        width: int = 1080,
        adblock: bool = False,
        automatic_consent: bool = False,
        **kwargs,
    ):
        from minet.browser import ThreadsafeBrowser

        self.browser = ThreadsafeBrowser(
            width=width,
            height=height,
            adblock=adblock,
            automatic_consent=automatic_consent,
        )

        super().__init__(
            max_workers,
            wait=wait,
            daemonic=daemonic,
            **kwargs,
        )

    def __enter__(self):
        self.browser.__enter__()
        return super().__enter__()

    def shutdown(self, wait=True) -> None:
        self.browser.stop()
        return super().shutdown(wait=wait)

    def run(
        self,
        iterator: Iterable[ItemType],
        fn: Callable[
            ["BrowserOrBrowserContext", HTTPWorkerPayload[ItemType]],
            Awaitable[ResultType],
        ],
        *,
        ordered: bool = False,
        key: Optional[Callable[[ItemType], Optional[str]]] = None,
        passthrough: bool = False,
        throttle: float = DEFAULT_THROTTLE,
        buffer_size: int = DEFAULT_IMAP_BUFFER_SIZE,
        domain_parallelism: int = DEFAULT_DOMAIN_PARALLELISM,
    ) -> Iterator[Tuple[ItemType, Optional[ResultType]]]:
        method = super().imap if ordered else super().imap_unordered

        def worker(
            payload: HTTPWorkerPayloadBase[ItemType],
        ) -> Tuple[ItemType, Optional[ResultType]]:
            if payload.url is None:
                return payload.item, None

            return payload.item, self.browser.run_in_browser_or_default_context(
                fn, cast(HTTPWorkerPayload[ItemType], payload)
            )

        imap = method(
            payloads_iter(iterator, key=key, passthrough=passthrough),
            worker,
            key=key_by_domain_name,
            parallelism=domain_parallelism,
            buffer_size=buffer_size,
            throttle=throttle,
        )

        yield from imap



from typing import List, Tuple, Optional

import re
import json
from html import unescape
from urllib.parse import unquote

from minet.scrape import WonderfulSoup
from minet.web import (
    create_pool_manager,
    create_request_retryer,
    request,
    retrying_method,
    Response,
)

from minet.youtube.types import YouTubeCaptionTrack, YouTubeCaptionLine
from minet.youtube.utils import ensure_video_id
from minet.youtube.exceptions import YouTubeInvalidVideoTargetError


CAPTION_TRACKS_RE = re.compile(r'"captionTracks":(\[.*?\])')


def select_caption_track(
    tracks: List[YouTubeCaptionTrack], langs: List[str], strict: bool = True
) -> Optional[YouTubeCaptionTrack]:
    def key(t):
        try:
            return (langs.index(t.lang), int(t.generated), t.lang)
        except ValueError:
            return (len(langs), int(t.generated), t.lang)

    best = min(tracks, key=key)

    if strict and best.lang not in langs:
        return None

    return best


class YouTubeScraper:
    def __init__(self):
        self.pool_manager = create_pool_manager()
        self.retryer = create_request_retryer()

    @retrying_method()
    def request(self, url, spoof_ua: bool = False) -> Response:
        return request(
            url,
            known_encoding="utf-8",
            pool_manager=self.pool_manager,
            spoof_ua=spoof_ua,
        )

    # Hats off to: https://github.com/algolia/youtube-captions-scraper
    def get_caption_tracks(self, video_id: str) -> List[YouTubeCaptionTrack]:
        # First we try to retrieve it from video info
        url = "https://www.youtube.com/watch?v=%s" % video_id

        response = self.request(url)

        data = unquote(response.text())

        m = CAPTION_TRACKS_RE.search(data)

        if m is not None:
            data = json.loads("{" + m.group(0) + "}")["captionTracks"]

            return [
                YouTubeCaptionTrack(
                    item["languageCode"], item["baseUrl"], item.get("kind") == "asr"
                )
                for item in data
            ]

        return []

    def get_video_captions(
        self, video_target: str, langs: List[str] = ["en"]
    ) -> Optional[Tuple[YouTubeCaptionTrack, List[YouTubeCaptionLine]]]:
        video_id = ensure_video_id(video_target)

        if video_id is None:
            raise YouTubeInvalidVideoTargetError

        tracks = self.get_caption_tracks(video_id)

        if not tracks:
            return

        best_track = select_caption_track(tracks, langs=langs)

        if best_track is None:
            return

        response = self.request(best_track.url)

        soup = WonderfulSoup(response.text(), "xml")

        captions = []

        for item in soup.select("text"):
            captions.append(
                YouTubeCaptionLine(
                    float(item["start"]),
                    float(item["dur"]),
                    unescape(item.get_text().strip()),
                )
            )

        return best_track, captions

    def get_channel_id(self, channel_url: str) -> Optional[str]:
        response = self.request(
            channel_url,
            spoof_ua=True,
        )

        soup = WonderfulSoup(response.text(), "lxml")
        tag = soup.force_select_one("meta[itemprop=identifier]")

        if tag:
            return tag.get("content")

        return None



# =============================================================================
# Minet Telegram Constants
# =============================================================================
#
# General constants used throughout the Telegram functions.
#
TELEGRAM_URL = "https://t.me/"
TELEGRAM_DEFAULT_THROTTLE = 0.5
TELEGRAM_INFOS_CSV_HEADERS = [
    "title",
    "name",
    "link",
    "img",
    "description",
    "nb_subscribers",
    "nb_photos",
    "nb_videos",
    "nb_links",
]
TELEGRAM_MESSAGES_CSV_HEADERS = [
    "link_to_message",
    "could_be_displayed",
    "user",
    "user_link",
    "user_img",
    "text",
    "nb_hashtags",
    "hashtags",
    "is_reply_img",
    "is_reply_user",
    "is_reply_text",
    "stickers",
    "nb_photos",
    "photos",
    "nb_videos",
    "videos",
    "videos_times",
    "nb_links",
    "links",
    "link_img",
    "link_site",
    "link_title",
    "link_description",
    "views",
    "datetime",
    "edited",
]



# =============================================================================
# Minet Crawl CLI Action
# =============================================================================
#
# Logic of the crawl action.
#
from typing import Optional, List, Callable, Any, Mapping, Union, cast

import os
import casanova
import casanova.ndjson as ndjson
from inspect import isclass
from os.path import join, dirname
from ebbe.decorators import with_defer

from minet.utils import import_target
from minet.fs import FilenameBuilder
from minet.exceptions import (
    GenericModuleNotFoundError,
    TargetInGenericModuleNotFoundError,
)
from minet.cli.exceptions import FatalError
from minet.crawl import (
    Crawler,
    CrawlResult,
    SuccessfulCrawlResult,
    SpiderDeclaration,
    Spider,
    BasicSpider,
)
from minet.crawl.exceptions import CrawlerAlreadyFinishedError
from minet.cli.console import console
from minet.cli.loading_bar import LoadingBar
from minet.cli.utils import (
    with_loading_bar,
    with_ctrl_c_warning,
    track_crawler_state_with_loading_bar,
)


# NOTE: here are the possible file structure for scraped data:
#
# 1. singular spider, one group:
#    - data.csv
# 2. singular spider, multiple groups:
#    - data/
#      - group1.csv
#      - group2.csv
# 3. plural spider, one group:
#    - data/
#      - spider1.csv
#      - spider2.csv
# 4. plural spider, multiple groups
#    - data/
#      - spider1/
#        - group1.csv
class DataWriter:
    def __init__(
        self, base_dir: str, crawler: Crawler, resume: bool = False, format: str = "csv"
    ):
        self.handles = {}
        self.resume = resume
        self.singular = crawler.singular
        self.format = format
        self.dir = base_dir

        if self.singular:
            self.__add_file(None, "data", crawler.get_spider())
        else:
            for name, spider in crawler.spiders():
                self.__add_file(name, join("data", name), spider)

    def __add_file(self, name: Optional[str], path: str, spider):
        path += "." + self.format

        path = join(self.dir, path)
        directory = dirname(path)
        os.makedirs(directory, exist_ok=True)

        f = (
            casanova.BasicResumer(path, encoding="utf-8")
            if self.resume
            else open(path, "w", encoding="utf-8")
        )

        if self.format == "csv":
            # TODO: ability to pass fieldnames? from spider?
            w = casanova.InferringWriter(f, add=["job_id"])
        elif self.format == "jsonl" or self.format == "ndjson":
            w = ndjson.writer(f)
        else:
            raise NotImplementedError('unknown format "%s"' % self.format)

        self.handles[name] = {"file": f, "writer": w, "spider": spider}

    def __unpack_result(self, result: SuccessfulCrawlResult, data):
        job_id = result.job.id

        if self.format == "csv":
            return (data, [job_id])

        return ({"job_id": job_id, "data": data},)

    def write(self, result: SuccessfulCrawlResult) -> None:
        handle = self.handles[result.spider]
        spider = handle["spider"]
        writer = handle["writer"]
        f = handle["file"]

        for item in spider.tabulate(result):
            writer.writerow(*self.__unpack_result(result, item))

        # NOTE: write is flushing to ensure atomicity as well as possible
        f.flush()

    def flush(self) -> None:
        for h in self.handles.values():
            h["file"].flush()

    def close(self) -> None:
        for h in self.handles.values():
            h["file"].close()


@with_defer()
@with_loading_bar(
    title="Crawling",
    unit="pages",
    stats=[
        {"name": "queued", "style": "info"},
        {"name": "doing", "style": "warning"},
        {"name": "done", "style": "success"},
    ],
)
@with_ctrl_c_warning
def crawl_action(
    cli_args,
    defer=None,
    loading_bar: Optional[LoadingBar] = None,
    target: Optional[Union[SpiderDeclaration, Callable[..., Crawler]]] = None,
    additional_job_fieldnames: Optional[List[str]] = None,
    format_job_row_addendum: Optional[Callable[[CrawlResult], List]] = None,
    result_callback: Optional[Callable[[Any, LoadingBar, CrawlResult], None]] = None,
):
    if (additional_job_fieldnames is not None and format_job_row_addendum is None) or (
        additional_job_fieldnames is None and format_job_row_addendum is not None
    ):
        raise TypeError("additional_job_fieldnames requires format_job_row_addendum")

    # NOTE: typing and decorators don't play well together
    assert defer is not None
    assert isinstance(loading_bar, LoadingBar)

    persistent_storage_path = join(cli_args.output_dir, "store")
    writer_root_directory = join(cli_args.output_dir, "pages")

    filename_builder = FilenameBuilder(cli_args.folder_strategy)

    def callback(self: Crawler, result: SuccessfulCrawlResult) -> Optional[str]:
        if not cli_args.write_files:
            return

        response = result.response
        filename = result.job.id

        path = filename_builder(
            url=response.end_url,
            filename=filename,
            compressed=cli_args.compress_on_disk,
            ext=response.ext,
        )

        self.write(path, response.body, compress=cli_args.compress_on_disk)

        return path

    # Scaffolding output directory
    os.makedirs(cli_args.output_dir, exist_ok=True)

    crawler_kwargs = {
        "persistent_storage_path": persistent_storage_path,
        "writer_root_directory": writer_root_directory,
        "resume": cli_args.resume,
        "wait": False,
        "daemonic": False,
    }

    # NOTE: the cli_args given to this action might lack some items
    # so that crawler factory are more convenient to use.
    cli_args_to_forward_to_crawler = [
        "throttle",
        "max_depth",
        "visit_urls_only_once",
        "normalized_url_cache",
        ("threads", "max_workers"),
        ("spoof_user_agent", "spoof_ua"),
        "insecure",
        "domain_parallelism",
        ("processes", "process_pool_workers"),
        "timeout",
        "stateful_redirects",
        ("pycurl", "use_pycurl"),
        ("compress_transfer", "compressed"),
        "sqlar",
    ]

    for arg in cli_args_to_forward_to_crawler:
        if isinstance(arg, tuple):
            cli_arg_name, crawler_arg_name = arg
        else:
            cli_arg_name = arg
            crawler_arg_name = arg

        if hasattr(cli_args, cli_arg_name):
            crawler_kwargs[crawler_arg_name] = getattr(cli_args, cli_arg_name)

    retries = getattr(cli_args, "retries", 0)

    if retries:
        crawler_kwargs["retry"] = True
        crawler_kwargs["retryer_kwargs"] = {
            "retry_on_timeout": True,
            "max_attempts": 1 + retries,
        }

    if target is None:
        if cli_args.module is not None:
            try:
                target = import_target(cli_args.module, "spider")
            except GenericModuleNotFoundError:
                raise FatalError(
                    [
                        "Could not import %s!" % cli_args.module,
                        "Are you sure the module exists?",
                    ]
                )
            except TargetInGenericModuleNotFoundError as e:
                raise FatalError(
                    [
                        "Could not find the %s target in the %s module!"
                        % (e.name, e.path),
                        "Are you sure this class/function/variable exists in the module?",
                    ]
                )
        else:
            target = BasicSpider()

    # NOTE: target can be:
    #   - a crawler factory function
    #   - a spider class
    #   - a spider instance
    #   - a dict of spider instances
    #   - a simple callable
    try:
        if not cli_args.factory:
            # Is target a Spider class?
            if isclass(target) and issubclass(target, Spider):
                target = target()
            else:
                # NOTE: at that point, target can be:
                #   - a Spider instance
                #   - a dict of Spider instances
                #   - a callable
                valid_spiders_dict = isinstance(target, Mapping) and all(
                    isinstance(v, Spider) for v in target.values()
                )

                # TODO: inspect arity to weed out potential footguns
                if (
                    not valid_spiders_dict
                    and not isinstance(target, Spider)
                    and not callable(target)
                ):
                    # TODO: explain further
                    raise FatalError("Invalid crawling target!")

            # NOTE: target IS a spider declaration
            target = cast(SpiderDeclaration, target)

            crawler = Crawler(target, **crawler_kwargs)

        else:
            if not callable(target):
                raise FatalError("Factory should be callable!")

            # NOTE: target is a crawler factory
            target = cast(Callable[..., Crawler], target)

            crawler = target(**crawler_kwargs)

            if not isinstance(crawler, Crawler):
                raise FatalError("Factory did not return a crawler!")

    except CrawlerAlreadyFinishedError:
        loading_bar.erase()
        raise FatalError("[error]Crawler has already finished!")

    # Jobs output
    jobs_output_path = join(cli_args.output_dir, "jobs.csv")
    jobs_output = (
        casanova.BasicResumer(jobs_output_path, encoding="utf-8")
        if cli_args.resume
        else open(jobs_output_path, "w", encoding="utf-8")
    )
    jobs_fieldnames = CrawlResult.fieldnames(singular=crawler.singular)

    if cli_args.write_files:
        jobs_fieldnames += ["path"]

    if additional_job_fieldnames is not None:
        jobs_fieldnames += additional_job_fieldnames

    jobs_writer = casanova.Writer(jobs_output, fieldnames=jobs_fieldnames)
    defer(jobs_output.close)

    with crawler:
        # Enqueuing extraneous start jobs?
        if crawler.resuming:
            loading_bar.print("[log.time]Crawler will now resume…")
        elif getattr(cli_args, "input", None):
            for url in casanova.reader(cli_args.input).cells(cli_args.column):
                crawler.enqueue(url)  # type: ignore

        data_writer = None

        if cli_args.write_data:
            data_writer = DataWriter(
                cli_args.output_dir,
                crawler,
                resume=cli_args.resume,
                format=getattr(cli_args, "format", "csv"),
            )
            defer(data_writer.close)

        if not crawler.resuming and len(crawler) == 0:
            raise FatalError(
                [
                    "Started a crawler without any jobs.",
                    "This can happen if the command was given no start url nor -i/--input flag.",
                    "You can also implement start urls/targets on your spiders themselves if required.",
                    "Or maybe you forgot to --resume?",
                ],
                warning=True,
            )

        track_crawler_state_with_loading_bar(loading_bar, crawler.state)

        # Running crawler
        for result, result_path in crawler.crawl(callback=callback):
            with loading_bar.step():
                if cli_args.verbose:
                    console.print(result, highlight=True)

                if result_callback is not None:
                    result_callback(cli_args, loading_bar, result)

                job_row = result.as_csv_row()

                if cli_args.write_files:
                    job_row += [result_path]

                if format_job_row_addendum is not None:
                    job_row += format_job_row_addendum(result)

                jobs_writer.writerow(job_row)

                # Flushing to avoid sync issues as well as possible
                jobs_output.flush()

                if result.error is not None:
                    loading_bar.inc_stat(result.error_code, style="error")
                    continue

                if data_writer is not None:
                    data_writer.write(result)


action = crawl_action



# =============================================================================
# Minet CrowdTangle Posts CLI Action
# =============================================================================
#
# Logic of the `ct posts` action.
#
from minet.crowdtangle.constants import CROWDTANGLE_POST_CSV_HEADERS
from minet.cli.crowdtangle.utils import make_paginated_action

action = make_paginated_action(
    method_name="posts", item_name="posts", csv_headers=CROWDTANGLE_POST_CSV_HEADERS
)



from typing import Union, Optional, List, Dict, Any

from bs4 import BeautifulSoup, SoupStrainer
from casanova import CSVSerializer

from minet.scrape.soup import WonderfulSoup
from minet.scrape.analysis import ScraperAnalysisOutputType
from minet.scrape.exceptions import ScraperNotTabularError

AnyScrapableTarget = Union[str, WonderfulSoup, BeautifulSoup]


class ScraperBase(object):
    fieldnames: Optional[List[str]]
    plural: bool
    output_type: ScraperAnalysisOutputType
    serializer: CSVSerializer
    strainer: Optional[SoupStrainer]

    @property
    def singular(self) -> bool:
        return not self.plural

    def __repr__(self):
        return "<{name} plural={plural} output_type={output_type} strain={strain} fieldnames={fieldnames!r}>".format(
            name=self.__class__.__name__,
            plural=self.plural,
            strain=getattr(self.strainer, "css", None) if self.strainer else None,
            output_type=self.output_type,
            fieldnames=self.fieldnames,
        )

    def __call__(self, html: AnyScrapableTarget, context: Optional[Dict] = None) -> Any:
        raise NotImplementedError

    def as_csv_rows(
        self,
        html: AnyScrapableTarget,
        context: Optional[Dict] = None,
        plural_separator="|",
    ):
        if self.fieldnames is None:
            raise ScraperNotTabularError

        def generator():
            result = self.__call__(html, context=context)

            if result is None:
                return

            if not self.plural:
                result = [result]

            for item in result:
                if isinstance(item, dict):
                    assert self.fieldnames

                    item = self.serializer.serialize_dict_row(
                        item, self.fieldnames, plural_separator=plural_separator
                    )
                else:
                    item = [self.serializer(item, plural_separator=plural_separator)]

                yield item

        return generator()

    def as_csv_dict_rows(
        self,
        html: AnyScrapableTarget,
        context: Optional[Dict] = None,
        plural_separator="|",
    ):
        if self.fieldnames is None:
            raise ScraperNotTabularError

        def generator():
            result = self.__call__(html, context=context)

            if result is None:
                return

            if not self.plural:
                result = [result]

            for item in result:
                if isinstance(item, dict):
                    for k, v in item.items():
                        item[k] = self.serializer(v, plural_separator=plural_separator)
                else:
                    item = {
                        "value": self.serializer(
                            item, plural_separator=plural_separator
                        )
                    }

                yield item

        return generator()

    def as_records(self, html: AnyScrapableTarget, context: Optional[Dict] = None):
        result = self.__call__(html, context=context)

        if result is None:
            return

        if not self.plural:
            yield result
            return

        yield from result



from typing import Callable, TypeVar, Awaitable, Set, Optional, Union
from minet.types import Literal, Concatenate, ParamSpec

import os
import asyncio
import platform
from shutil import rmtree
from concurrent.futures import Future
from threading import Thread, Event, Lock
from playwright.async_api import async_playwright, Browser, BrowserContext

from minet.exceptions import UnknownBrowserError
from minet.browser.plawright_shim import install_browser
from minet.browser.utils import get_browsers_path, get_temp_persistent_context_path
from minet.browser.extensions import get_extension_path, ensure_extension_is_downloaded

UNIX = "windows" not in platform.system().lower()
LTE_PY37 = platform.python_version_tuple()[:2] <= ("3", "7")
SUPPORTED_BROWSERS = ("chromium", "firefox")


BrowserName = Literal["chromium", "firefox"]
T = TypeVar("T")
P = ParamSpec("P")
BrowserOrBrowserContext = Union[Browser, BrowserContext]
BrowserCallable = Callable[Concatenate[Browser, P], Awaitable[T]]
BrowserContextCallable = Callable[Concatenate[BrowserContext, P], Awaitable[T]]
BrowserOrBrowserContextCallable = Callable[
    Concatenate[BrowserOrBrowserContext, P], Awaitable[T]
]

# TODO: introduce stealth back
# TODO: abstract screenshot function
# TODO: helper for stealth page, monkey patch some function?


class ThreadsafeBrowser:
    def __init__(
        self,
        browser: BrowserName = "chromium",
        headless: bool = True,
        automatic_consent: bool = False,
        adblock: bool = False,
        width: int = 1920,
        height: int = 1080,
    ) -> None:
        if browser not in SUPPORTED_BROWSERS:
            raise UnknownBrowserError(browser)

        self.extensions = []

        if automatic_consent:
            self.extensions.append("i-still-dont-care-about-cookies")

        if adblock:
            self.extensions.append("ublock-origin")

        # NOTE: doing this in constructor to avoid threading issues down the line
        for ext in self.extensions:
            ensure_extension_is_downloaded(ext)

        self.requires_extensions = bool(self.extensions)
        self.persistent_user_data_dir = None
        self.persistent = self.requires_extensions

        if self.requires_extensions and browser != "chromium":
            raise TypeError("adblock and automatic_consent only work with chromium")

        # NOTE: on unix python 3.7, child watching does not
        # work properly when asyncio is not running from the main thread
        if UNIX and LTE_PY37:
            from minet.__future__.threaded_child_watcher import ThreadedChildWatcher

            asyncio.set_child_watcher(ThreadedChildWatcher())

        self.browser_name: str = browser
        self.browser: Optional[Browser] = None
        self.default_browser_context: Optional[BrowserContext] = None
        self.headless = headless
        self.width = width
        self.height = height

        os.environ.setdefault("PLAYWRIGHT_BROWSERS_PATH", get_browsers_path())

        install_browser(self.browser_name)

        self.loop = asyncio.new_event_loop()
        self.start_event = Event()
        self.thread = Thread(
            name="Thread-browser-%i" % id(self), target=self.__thread_worker
        )

        self.running_futures: Set[Future] = set()
        self.running_futures_lock = Lock()

        # Starting loop thread
        self.thread.start()
        self.start_event.wait()

    async def __start_playwright(self) -> None:
        self.playwright = await async_playwright().start()

        if self.browser_name == "chromium":
            browser_type = self.playwright.chromium
        elif self.browser_name == "firefox":
            browser_type = self.playwright.firefox
        else:
            raise TypeError("unsupported browser")

        if self.requires_extensions:
            args = []
            disable_extensions_except = []

            for ext in self.extensions:
                p = get_extension_path(ext)
                disable_extensions_except.append(p)
                args.append(f"--load-extension={p}")

            args.append(
                f"--disable-extensions-except={','.join(disable_extensions_except)}"
            )

            if self.headless:
                args.append("--headless=new")

            user_data_dir = get_temp_persistent_context_path()
            rmtree(user_data_dir, ignore_errors=True)
            self.persistent_user_data_dir = user_data_dir

            self.default_browser_context = await browser_type.launch_persistent_context(
                user_data_dir,
                headless=False,
                args=args,
                viewport={"width": self.width, "height": self.height},
                locale="en-US",
            )

        else:
            self.browser = await browser_type.launch(headless=self.headless)
            self.default_browser_context = await self.browser.new_context(
                viewport={"width": self.width, "height": self.height}, locale="en-US"
            )

    async def __stop_playwright(self) -> None:
        # NOTE: we need to make sure those were actually launched, in
        # case of a nasty race condition

        if self.default_browser_context is not None:
            await self.default_browser_context.close()

        if self.browser is not None:
            await self.browser.close()

        # NOTE: this hangs without the proper child watcher
        await self.playwright.stop()

    def stop(self) -> None:
        # NOTE: if we don't do this and some job sent
        # to a threadpool executor raises and triggers
        # the closing of the playwright driver, then a
        # catastrophic chain reaction of exception will
        # make some jobs hang up and block indefinitely
        # which will cause a deadlock.
        with self.running_futures_lock:
            for fut in self.running_futures:
                if not fut.done():
                    fut.cancel()

        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.stop()

    def __thread_worker(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self.__start_playwright())
        self.start_event.set()

        # NOTE: we are now ready to accept tasks
        try:
            self.loop.run_forever()
        finally:
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())

        self.loop.run_until_complete(self.__stop_playwright())

    def __handle_future(self, future: Future):
        with self.running_futures_lock:
            self.running_futures.add(future)

        try:
            return future.result()
        finally:
            with self.running_futures_lock:
                self.running_futures.remove(future)

    def run(self, fn: BrowserCallable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        if self.browser is None:
            raise TypeError("cannot run browser-level callable with persistent context")

        future = asyncio.run_coroutine_threadsafe(
            fn(self.browser, *args, **kwargs), self.loop
        )

        return self.__handle_future(future)

    def run_in_default_context(
        self, fn: BrowserContextCallable[P, T], *args: P.args, **kwargs: P.kwargs
    ) -> T:
        assert self.default_browser_context is not None

        future = asyncio.run_coroutine_threadsafe(
            fn(self.default_browser_context, *args, **kwargs), self.loop
        )

        return self.__handle_future(future)

    def run_in_browser_or_default_context(
        self,
        fn: BrowserOrBrowserContextCallable[P, T],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        target = self.browser

        if target is None:
            target = self.default_browser_context

        assert target is not None

        future = asyncio.run_coroutine_threadsafe(
            fn(target, *args, **kwargs), self.loop
        )

        return self.__handle_future(future)



from typing import TypeVar, Union, Generic, Optional, List

import json
from nanoid import generate
from ebbe import format_repr
from functools import partial

from minet.web import Response
from minet.encodings import normalize_encoding
from minet.serialization import serialize_error_as_slug

CrawlJobDataType = TypeVar("CrawlJobDataType")
CrawlResultDataType = TypeVar("CrawlResultDataType")


class CrawlTarget(Generic[CrawlJobDataType]):
    """
    This class represent a crawl target, i.e. at least a URL, and should be
    used by end-users when they need to specify what they want to crawl more
    precisely (i.e. if they need to switch spiders, have a custom depth or
    rely on some necessary data that will be passed along).
    """

    __slots__ = ("url", "depth", "spider", "priority", "data")

    url: str
    depth: Optional[int]
    spider: Optional[str]
    priority: int
    data: Optional[CrawlJobDataType]

    def __init__(
        self,
        url: str,
        depth: Optional[int] = None,
        spider: Optional[str] = None,
        priority: int = 0,
        data: Optional[CrawlJobDataType] = None,
    ) -> None:
        if not isinstance(url, str):
            raise TypeError("url should be a string")

        self.url = url

        if depth is not None and not isinstance(depth, int):
            raise TypeError("depth should be an int")

        self.depth = depth

        if spider is not None and not isinstance(spider, str):
            raise TypeError("spider should be a string")

        if not isinstance(priority, int):
            raise TypeError("depth should be an int")

        self.spider = spider
        self.priority = priority
        self.data = data

    def __eq__(self, other):
        if not isinstance(other, CrawlTarget):
            return False

        for k in self.__slots__:
            if getattr(self, k) != getattr(other, k):
                return False

        return True

    def __repr__(self):
        return format_repr(
            self,
            ("url", "spider", "depth", ("priority", self.priority or None), "data"),
            conditionals=("spider", "data", "depth", "priority"),
        )


UrlOrCrawlTarget = Union[str, CrawlTarget[CrawlJobDataType]]


generate_crawl_job_id = partial(
    generate,
    alphabet="0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
    size=12,
)


class CrawlJob(Generic[CrawlJobDataType]):
    """
    This class represents an actual job that will be queued by the crawler. It
    should not be created by users themselves as they are inferred from
    a CrawlTarget instead.

    This class should be pickle-safe as it will be serialized by persistent
    queues.
    """

    __slots__ = ("id", "url", "group", "depth", "spider", "priority", "data", "parent")

    id: str
    url: str
    group: Optional[str]
    depth: int
    spider: Optional[str]
    priority: int
    data: Optional[CrawlJobDataType]
    parent: Optional[str]

    # TODO: we should add headers, method, cookies and such here in the future,
    # but for now, a request_args override can do the trick based on job data

    @classmethod
    def fieldnames(cls):
        return [slot for slot in cls.__slots__ if not slot.startswith("_")]

    def __init__(
        self,
        url: str,
        id: Optional[str] = None,
        group: Optional[str] = None,
        depth: Optional[int] = None,
        spider: Optional[str] = None,
        priority: int = 0,
        data: Optional[CrawlJobDataType] = None,
        parent: Optional[str] = None,
    ):
        self.id = id if id is not None else generate_crawl_job_id()
        self.url = url
        self.group = group
        self.depth = depth if depth is not None else 0
        self.spider = spider
        self.priority = priority
        self.data = data
        self.parent = parent

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other):
        if not isinstance(other, CrawlJob):
            return False

        for k in self.__slots__:
            if getattr(self, k) != getattr(other, k):
                return False

        return True

    def __tabulate(self, serialize_data: bool = False):
        return (
            self.id,
            self.url,
            self.group,
            self.depth,
            self.spider,
            self.priority,
            self.data
            if not serialize_data
            else (
                json.dumps(self.data, ensure_ascii=False)
                if self.data is not None
                else None
            ),
            self.parent,
        )

    def __getstate__(self):
        return self.__tabulate()

    def __csv_row__(self):
        return self.__tabulate(serialize_data=True)

    def __setstate__(self, state):
        (
            self.id,
            self.url,
            self.group,
            self.depth,
            self.spider,
            self.priority,
            self.data,
            self.parent,
        ) = state

    def __repr__(self):
        return format_repr(
            self,
            (
                "id",
                "url",
                "depth",
                "spider",
                ("priority", self.priority or None),
                "data",
                "parent",
            ),
            conditionals=("data", "spider", "parent", "group", "priority"),
        )


class CrawlResult(Generic[CrawlJobDataType, CrawlResultDataType]):
    __slots__ = ("job", "data", "error", "response", "degree")

    FIELDNAMES = [
        "id",
        "parent",
        "spider",
        "depth",
        "url",
        "resolved_url",
        "error",
        "status",
        "mimetype",
        "degree",
        "body_size",
        "encoding",
    ]

    job: CrawlJob[CrawlJobDataType]
    data: Optional[CrawlResultDataType]
    error: Optional[Exception]
    response: Optional[Response]
    degree: int

    @classmethod
    def fieldnames(cls, singular: bool = False) -> List[str]:
        return (
            cls.FIELDNAMES
            if not singular
            else [f for f in cls.FIELDNAMES if f != "spider"]
        )

    def __init__(self, job: CrawlJob[CrawlJobDataType]):
        self.job = job
        self.data = None
        self.error = None
        self.response = None
        self.degree = 0

    @property
    def url(self) -> str:
        return self.job.url

    @property
    def depth(self) -> int:
        return self.job.depth

    @property
    def spider(self) -> Optional[str]:
        return self.job.spider

    @property
    def error_code(self) -> Optional[str]:
        return serialize_error_as_slug(self.error) if self.error else None

    def __csv_row__(self):
        job = self.job

        row = [job.id, job.parent]

        if job.spider is not None:
            row.append(job.spider)

        row.extend(
            [
                job.depth,
                job.url,
                self.response.end_url if self.response else None,
                self.error_code,
                self.response.status if self.response else None,
                self.response.mimetype if self.response else None,
                self.degree,
                len(self.response) if self.response else None,
                self.response.encoding if self.response else None,
            ]
        )

        return row

    def as_csv_row(self):
        return self.__csv_row__()

    def __repr__(self) -> str:
        encoding = self.response.encoding if self.response else None

        if encoding and normalize_encoding(encoding) == "utf8":
            encoding = None

        return format_repr(
            self,
            attributes=(
                (
                    "url",
                    "depth",
                    "spider",
                    ("error", self.error_code),
                    ("status", self.response.status if self.response else None),
                    ("degree", self.degree if self.error is None else None),
                    (
                        "dtype",
                        type(self.data).__name__ if self.data is not None else None,
                    ),
                    ("size", self.response.human_size if self.response else None),
                    ("mimetype", self.response.mimetype if self.response else None),
                    ("encoding", encoding),
                )
            ),
            conditionals=(
                "spider",
                "error",
                "status",
                "degree",
                "dtype",
                "size",
                "mimetype",
                "encoding",
            ),
        )


class ErroredCrawlResult(CrawlResult[CrawlJobDataType, None]):
    job: CrawlJob[CrawlJobDataType]
    data: None
    error: Exception
    response: Optional[Response]
    degree: int

    def __init__(
        self,
        job: CrawlJob[CrawlJobDataType],
        error: Exception,
        response: Optional[Response] = None,
    ):
        self.job = job
        self.data = None
        self.error = error
        self.response = response
        self.degree = 0

    @property
    def error_code(self) -> str:
        return serialize_error_as_slug(self.error)


class SuccessfulCrawlResult(CrawlResult[CrawlJobDataType, CrawlResultDataType]):
    job: CrawlJob[CrawlJobDataType]
    data: CrawlResultDataType
    error: None
    response: Response
    degree: int

    def __init__(
        self,
        job: CrawlJob[CrawlJobDataType],
        response: Response,
        data: CrawlResultDataType,
        degree: int,
    ):
        self.job = job
        self.data = data
        self.error = None
        self.response = response
        self.degree = degree

    @property
    def error_code(self) -> None:
        return None


AnyCrawlResult = Union[
    ErroredCrawlResult[CrawlJobDataType],
    SuccessfulCrawlResult[CrawlJobDataType, CrawlResultDataType],
]



# =============================================================================
# Minet Scrape CLI Action
# =============================================================================
#
# Logic of the scrape action.
#
from typing import Optional, Any, List, Dict, Iterator

import casanova
import casanova.ndjson as ndjson
from dataclasses import dataclass
from itertools import count
from os.path import basename, isdir

from minet.scrape import Scraper
from minet.scrape.typical import TYPICAL_SCRAPERS
from minet.scrape.types import ScraperBase
from minet.multiprocessing import LazyPool
from minet.exceptions import (
    DefinitionInvalidFormatError,
)
from minet.scrape.exceptions import (
    InvalidScraperError,
    CSSSelectorTooComplex,
    ScraperEvalError,
    ScraperEvalTypeError,
    ScraperEvalNoneError,
)
from minet.cli.reporters import (
    report_scraper_validation_errors,
    report_scraper_evaluation_error,
)
from minet.serialization import serialize_error_as_slug
from minet.fs import read_potentially_gzipped_path
from minet.cli.console import console
from minet.cli.loading_bar import LoadingBar
from minet.cli.utils import create_fetch_like_report_iterator, FetchReportLikeItem
from minet.cli.exceptions import FatalError


@dataclass
class ScrapeWorkerPayload:
    id: int
    row: List[str]
    path: Optional[str]
    encoding: Optional[str]
    text: Optional[str]
    url: Optional[str]


@dataclass
class ScrapeResult:
    id: int
    error: Optional[Exception] = None
    items: Optional[Any] = None


SCRAPER: Optional[ScraperBase] = None
FORMAT: Optional[str] = None
PLURAL_SEPARATOR: Optional[str] = None
HEADERS: Optional[casanova.headers] = None


def init_process(options):
    global SCRAPER
    global FORMAT
    global PLURAL_SEPARATOR
    global HEADERS

    if options["name"] is not None:
        SCRAPER = TYPICAL_SCRAPERS[options["name"]]()
    else:
        SCRAPER = Scraper(options["definition"], strain=options["strain"])

    FORMAT = options["format"]
    PLURAL_SEPARATOR = options["plural_separator"]
    HEADERS = casanova.headers(options["fieldnames"])


def worker(payload: ScrapeWorkerPayload) -> ScrapeResult:
    assert SCRAPER is not None
    assert PLURAL_SEPARATOR is not None
    assert HEADERS is not None

    text = payload.text

    result = ScrapeResult(id=payload.id)

    # Reading file
    if text is None:
        try:
            text = read_potentially_gzipped_path(
                payload.path, encoding=payload.encoding, fallback_encoding="utf-8"
            )
        except (FileNotFoundError, UnicodeDecodeError) as e:
            result.error = e
            return result

    # Building context
    row = HEADERS.wrap(payload.row, transient=True)
    context: Dict[str, Any] = {"row": row}

    if payload.url:
        context["url"] = payload.url

    if payload.path:
        context["path"] = payload.path
        context["basename"] = basename(payload.path)

    # Attempting to scrape
    if FORMAT == "csv":
        items = SCRAPER.as_csv_rows(
            text, context=context, plural_separator=PLURAL_SEPARATOR
        )
    else:
        items = SCRAPER.as_records(text, context=context)

    # NOTE: errors might be raised when we consume the generators created above
    try:
        items = list(items)
    except (ScraperEvalError, ScraperEvalTypeError, ScraperEvalNoneError) as error:
        result.error = error
        return result

    result.items = items

    return result


def action(cli_args):
    using_typical_scraper = False

    # Parsing scraper definition
    try:
        if cli_args.scraper in TYPICAL_SCRAPERS:
            using_typical_scraper = True
            scraper = TYPICAL_SCRAPERS[cli_args.scraper]()
        else:
            scraper = Scraper(cli_args.scraper, strain=cli_args.strain)

    except DefinitionInvalidFormatError:
        raise FatalError(
            ["Unknown scraper format!", "It should be a JSON or YAML file."]
        )

    except FileNotFoundError:
        raise FatalError("Could not find scraper file!")

    except InvalidScraperError as error:
        raise FatalError(
            [
                "Your scraper is invalid! You need to fix the following errors:\n",
                report_scraper_validation_errors(error.validation_errors),
            ]
        )

    except CSSSelectorTooComplex:
        raise FatalError(
            [
                "Your strainer's CSS selector [info]%s[/info] is too complex."
                % cli_args.strain,
                "You cannot use relations to create a strainer.",
                "Try to simplify the selector you passed to --strain.",
            ]
        )

    if scraper.fieldnames is None and cli_args.format == "csv":
        raise FatalError(
            [
                "Your scraper does not yield tabular data.",
                'Try changing it or setting --format to "jsonl".',
            ]
        )

    if cli_args.input_dir is not None and not isdir(cli_args.input_dir):
        raise FatalError(
            'Could not find the [cyan]-I/--input-dir "{}"[/cyan] directory!'.format(
                cli_args.input_dir
            )
        )

    if cli_args.format == "csv":
        assert scraper.fieldnames is not None

        output_fieldnames = scraper.fieldnames

        if cli_args.scraped_column_prefix is not None:
            output_fieldnames = [
                cli_args.scraped_column_prefix + h for h in output_fieldnames
            ]

        enricher = casanova.enricher(
            cli_args.input,
            cli_args.output,
            total=cli_args.total,
            select=cli_args.select,
            add=output_fieldnames,
        )
        reader = enricher

        def writerow(row, item):
            enricher.writerow(row, item)

    else:
        # TODO: casanova should probably expose some ndjson enricher
        reader = casanova.reader(cli_args.input, total=cli_args.total)
        writer = ndjson.writer(cli_args.output)

        def writerow(row, item):
            writer.writerow(item)

    with LoadingBar(
        "Scraping", unit="pages", total=reader.total if not cli_args.glob else None
    ) as loading_bar:
        items = create_fetch_like_report_iterator(cli_args, reader)

        worked_on: Dict[int, FetchReportLikeItem] = {}

        def payloads() -> Iterator[ScrapeWorkerPayload]:
            current_id = count()

            for item in items:
                # Items we cannot process
                if item.error is not None:
                    loading_bar.advance()
                    loading_bar.inc_stat(item.error, style="error")

                    # NOTE: we emit an empty line on error if scraper is singular
                    writerow(item.row, None)
                    continue

                item_id = next(current_id)

                worked_on[item_id] = item

                yield ScrapeWorkerPayload(
                    id=item_id,
                    encoding=item.encoding,
                    path=item.path,
                    text=item.text,
                    row=item.row,
                    url=item.url,
                )

        pool = LazyPool(
            cli_args.processes,
            initializer=init_process,
            initargs=(
                {
                    "name": cli_args.scraper if using_typical_scraper else None,
                    "definition": getattr(scraper, "definition", None),
                    "strain": cli_args.strain if not using_typical_scraper else None,
                    "format": cli_args.format,
                    "plural_separator": cli_args.plural_separator,
                    "fieldnames": reader.fieldnames,
                },
            ),
        )

        loading_bar.append_to_title(" (p=%i)" % pool.processes)

        warned_about_input_dir = False

        is_singular = scraper.singular

        with pool:
            for result in pool.imap(
                worker,
                payloads(),
                chunksize=cli_args.chunk_size,
                unordered=cli_args.unordered,
            ):
                with loading_bar.step():
                    original_item = worked_on.pop(result.id)

                    if result.error is not None:
                        # NOTE: we emit an empty line on error if scraper is singular
                        if is_singular:
                            writerow(original_item.row, None)

                        if isinstance(
                            result.error,
                            (
                                ScraperEvalError,
                                ScraperEvalTypeError,
                                ScraperEvalNoneError,
                            ),
                        ):
                            loading_bar.print(
                                report_scraper_evaluation_error(result.error), end=""
                            )
                        else:
                            error_code = serialize_error_as_slug(result.error)
                            loading_bar.inc_stat(error_code, style="error")

                            if (
                                not warned_about_input_dir
                                and error_code == "file-not-found"
                            ):
                                warned_about_input_dir = True

                                if cli_args.input_dir is None:
                                    console.warning(
                                        "did you forget to give [cyan]-I/--input-dir[/cyan]? "
                                    )
                                else:
                                    console.warning(
                                        'are you sure [cyan]-I/--input-dir "{}"[/cyan] is the correct path?'.format(
                                            cli_args.input_dir
                                        )
                                    )

                        continue

                    # No errors
                    assert result.items is not None
                    items = result.items

                    for item in items:
                        writerow(original_item.row, item)

                    # NOTE: we emit an empty line for singular scraper if no match occurred
                    if is_singular and not items:
                        writerow(original_item.row, None)

                    loading_bar.inc_stat(
                        "scraped-items", count=len(items), style="info"
                    )



# =============================================================================
# Minet BuzzSumo API Client
# =============================================================================
#
# BuzzSumo API client.
#
from json import JSONDecodeError
from urllib.parse import quote

from minet.rate_limiting import RateLimiterState, rate_limited_method
from minet.web import create_request_retryer, retrying_method, request

from minet.buzzsumo.formatters import format_article, format_exact_url
from minet.buzzsumo.exceptions import (
    BuzzSumoInvalidTokenError,
    BuzzSumoOutageError,
    BuzzSumoBadRequestError,
    BuzzSumoInvalidQueryError,
)

URL_TEMPLATE = "https://api.buzzsumo.com%s?api_key=%s"
MAXIMUM_PAGE_NB = 98


def construct_url(
    endpoint,
    token,
    begin_timestamp=None,
    end_timestamp=None,
    num_results=100,
    q=None,
    page=None,
):
    url = URL_TEMPLATE % (endpoint, token)

    url += "&num_results=%i" % num_results

    if begin_timestamp:
        url += "&begin_date=%s" % begin_timestamp
    if end_timestamp:
        url += "&end_date=%s" % (end_timestamp - 1)
    if q is not None:
        url += "&q=%s" % quote(q)
    if page is not None:
        url += "&page=%i" % page

    return url


def optimize_period_timestamps_wrt_nb_pages(
    period_timestamps, nb_pages, maximum_page_nb
):
    new_period_timestamps = period_timestamps

    if any(nb_page > maximum_page_nb for nb_page in nb_pages):
        for i in range(len(nb_pages)):
            if nb_pages[i] > maximum_page_nb:
                new_period_timestamps.append(
                    (period_timestamps[i] + period_timestamps[i + 1]) / 2
                )

        new_period_timestamps.sort()

    return period_timestamps


class BuzzSumoAPIClient(object):
    def __init__(self, token):
        self.token = token
        self.retryer = create_request_retryer(
            additional_exceptions=[BuzzSumoOutageError]
        )

        # 10 calls per ~12s
        self.rate_limiter_state = RateLimiterState(10, 12)

    @retrying_method()
    @rate_limited_method()
    def request(self, url):
        try:
            response = request(url, known_encoding="utf-8")
            data = response.json()
        except JSONDecodeError:
            raise BuzzSumoBadRequestError

        if response.status == 401:
            raise BuzzSumoInvalidTokenError

        if response.status == 406:
            raise BuzzSumoInvalidQueryError(data.get("error"), url=url, data=data)

        if response.status == 500:
            raise BuzzSumoBadRequestError

        if response.status > 500:
            raise BuzzSumoOutageError

        return response, data

    def limit(self):
        url = construct_url(
            "/search/articles.json", token=self.token, q="fake news", num_results=1
        )

        response, _ = self.request(url)

        return response.headers["X-RateLimit-Month-Remaining"]

    def domain_summary(self, domain, begin_timestamp, end_timestamp):
        url = construct_url(
            "/search/articles.json",
            token=self.token,
            q=domain,
            begin_timestamp=begin_timestamp,
            end_timestamp=end_timestamp,
        )

        _, data = self.request(url)

        return {
            "total_results": int(data["total_results"]),
            "total_pages": data["total_pages"],
        }

    def exact_url(self, search_url, begin_timestamp, end_timestamp):
        url = (
            construct_url(
                "/search/articles.json",
                token=self.token,
                q=search_url,
                begin_timestamp=begin_timestamp,
                end_timestamp=end_timestamp,
            )
            + "&exact_url=True"
        )

        _, data = self.request(url)

        if isinstance(data.get("results"), list) and len(data["results"]) == 1:
            match = data["results"][0]
            return format_exact_url(match)

    def __get_nb_pages_per_period_dates(self, domain, period_timestamps):
        nb_pages = []

        for i in range(len(period_timestamps) - 1):
            info = self.domain_summary(
                domain,
                begin_timestamp=period_timestamps[i],
                end_timestamp=period_timestamps[i + 1],
            )

            nb_pages.append(info["total_pages"])

        return nb_pages

    def domain_articles(self, domain, begin_timestamp, end_timestamp):
        # Here we optimize the periods used to request the API, because BuzzSumo
        # prevents us from getting more than 99 pages.
        maximum_page_nb = MAXIMUM_PAGE_NB
        period_timestamps = [begin_timestamp, end_timestamp]
        nb_pages = [1000]

        # This loop creates adapted time periods that all return less than 99 pages of results:
        while any(nb_page > maximum_page_nb for nb_page in nb_pages):
            # We ask how many pages are needed to get all the articles for the given periods:
            nb_pages = self.__get_nb_pages_per_period_dates(domain, period_timestamps)

            # If a given period gets more than 98 pages, this period is then cut down in half:
            period_timestamps = optimize_period_timestamps_wrt_nb_pages(
                period_timestamps, nb_pages, maximum_page_nb
            )

        # Now we get all the results for the optimized periods
        for i in range(len(period_timestamps) - 1):
            page = 0

            while True:
                url = construct_url(
                    "/search/articles.json",
                    token=self.token,
                    q=domain,
                    begin_timestamp=period_timestamps[i],
                    end_timestamp=period_timestamps[i + 1],
                    page=page,
                )

                _, data = self.request(url)

                if not data["results"]:
                    break

                for article in data["results"]:
                    yield format_article(article)

                page += 1



# =============================================================================
# Minet BuzzSumo Exact URL CLI Action
# =============================================================================
#
# Logic of the `bz exact-url` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.buzzsumo.utils import with_buzzsumo_fatal_errors
from minet.buzzsumo import BuzzSumoAPIClient
from minet.buzzsumo.constants import EXACT_URL_HEADERS


@with_buzzsumo_fatal_errors
@with_enricher_and_loading_bar(
    title="Retrieving url metadata", headers=EXACT_URL_HEADERS, unit="urls"
)
def action(cli_args, enricher, loading_bar):
    client = BuzzSumoAPIClient(cli_args.token)

    for row, url in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            data = client.exact_url(url, cli_args.begin_date, cli_args.end_date)
            if data:
                enricher.writerow(row, data.as_csv_row())
            else:
                enricher.writerow(row)



# =============================================================================
# Minet BuzzSumo Formatters
# =============================================================================
#
# Various formatters for BuzzSumo data.
#
from casanova import namedrecord

from minet.buzzsumo.constants import ARTICLES_CSV_HEADERS, EXACT_URL_HEADERS


BuzzSumoArticle = namedrecord(
    "BuzzSumoArticle",
    ARTICLES_CSV_HEADERS,
    plural=["article_amplifiers", "article_types"],
)


def format_article(data):
    keep = {}

    for k in ARTICLES_CSV_HEADERS:
        keep[k] = data[k]

    return BuzzSumoArticle(**keep)


BuzzSumoExactURL = namedrecord(
    "BuzzSumoExactURL",
    EXACT_URL_HEADERS,
    plural=["article_types"],
)


def format_exact_url(data):
    keep = {}

    for k in EXACT_URL_HEADERS:
        keep[k] = data.get(k)

    return BuzzSumoExactURL(**keep)



# Taken from playwright.__main__ so it is possible to run the command line
# tool directly from a python execution
import subprocess
from glob import iglob
from os.path import join, isdir

from playwright._impl._driver import compute_driver_executable, get_driver_env

from minet.loggers import downloaders_logger
from minet.browser.utils import get_browsers_path


def run_playwright(*args: str) -> int:
    env = get_driver_env()
    env.setdefault("PLAYWRIGHT_BROWSERS_PATH", get_browsers_path())

    driver_executable = compute_driver_executable()
    completed_process = subprocess.run(
        [str(driver_executable), *args], env=env, stdout=subprocess.DEVNULL
    )

    return completed_process.returncode


def install_browser(name: str) -> None:
    path = join(get_browsers_path(), name + "-*")

    if any(isdir(p) for p in iglob(path)):
        return

    downloaders_logger.info(
        "Installing browser: %s (this can take a little while, hang tight)" % name,
        extra={"namespace": "browser", "target": name},
    )

    run_playwright("install", name)



from typing import Optional, List, Dict, Any

from dataclasses import dataclass, field
from casanova import TabularRecord
from trafilatura.core import bare_extraction

from minet.exceptions import TrafilaturaError
from minet.encodings import fix_surrogates


def normalize_plural_trafilatura_item(meta: Dict[str, Any], key: str) -> List[str]:
    l = meta.get(key, []) or []

    items = []

    if not l:
        return items

    for item in l:
        if isinstance(item, dict):
            item = item.get("name")

            if item is None:
                continue

        if isinstance(item, (int, float)):
            item = str(item)

        if not isinstance(item, str):
            continue

        for subitem in item.split(","):
            subitem = subitem.strip()

            if subitem:
                items.append(subitem)

    # NOTE: trafilatura metadata extraction seem to produce invalid utf-8
    # sometimes, so we fix it here. We might need to fix this elsewhere
    # at some point also.
    items = [fix_surrogates(item) for item in items]

    return items


@dataclass
class TrafilaturaResult(TabularRecord):
    canonical_url: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    content: Optional[str] = None
    comments: Optional[str] = None
    author: Optional[str] = None
    categories: List[str] = field(default_factory=lambda: [])
    tags: List[str] = field(default_factory=lambda: [])
    date: Optional[str] = None
    sitename: Optional[str] = None

    @classmethod
    def from_bare_extraction(cls, data: Dict) -> "TrafilaturaResult":
        return TrafilaturaResult(
            canonical_url=data.get("url"),
            title=data.get("title"),
            description=data.get("description"),
            content=data.get("text"),
            comments=data.get("comments"),
            author=data.get("author"),
            categories=normalize_plural_trafilatura_item(data, "categories"),
            tags=normalize_plural_trafilatura_item(data, "tags"),
            date=data.get("date"),
            sitename=data.get("sitename"),
        )

    def blurb(self, fields=None) -> str:
        selection = fields if fields else self.fieldnames()
        items = []
        for s in selection:
            attr = getattr(self, s, "")
            if isinstance(attr, list):
                items.append(" ".join(attr))
            elif attr:
                items.append(attr)

        return "\n".join(items)


def extract(text: str) -> Optional[TrafilaturaResult]:
    # Attempting extraction
    try:
        # https://trafilatura.readthedocs.io/en/latest/corefunctions.html
        trafilatura_bare_result = bare_extraction(text)
    except Exception as e:
        raise TrafilaturaError(reason=e)

    if trafilatura_bare_result is None:
        return

    return TrafilaturaResult.from_bare_extraction(trafilatura_bare_result)



# =============================================================================
# Minet Instagram Constants
# =============================================================================
#
# General constants used throughout the Instagram functions.
#
from urllib3 import Timeout

INSTAGRAM_URL = "https://www.instagram.com"
INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 2)
INSTAGRAM_MEDIA_TYPE = {1: "GraphImage", 2: "GraphVideo", 8: "GraphSidecar"}
INSTAGRAM_DEFAULT_THROTTLE = 30.0
INSTAGRAM_MAX_RANDOM_ADDENDUM = 10.0
INSTAGRAM_NB_REQUEST_LITTLE_WAIT = 10
INSTAGRAM_DEFAULT_THROTTLE_LITTLE_WAIT = 300.0
INSTAGRAM_MAX_RANDOM_ADDENDUM_LITTLE_WAIT = 300.0
INSTAGRAM_NB_REQUEST_BIG_WAIT = 2900
INSTAGRAM_DEFAULT_THROTTLE_BIG_WAIT = 1800.0
INSTAGRAM_MAX_RANDOM_ADDENDUM_BIG_WAIT = 1800.0
INSTAGRAM_MIN_TIME_RETRYER = 600.0

INSTAGRAM_COMMENT_CSV_HEADERS = [
    "id",
    "index",
    "parent_id",
    "child_index",
    "text",
    "hashtags",
    "mentioned_names",
    "has_translation",
    "comment_like_count",
    "is_liked_by_media_owner",
    "child_comment_count",
    "created_at_utc",
    "utc_time",
    "username",
    "user_id",
    "user_full_name",
    "user_is_private",
    "user_is_mentionable",
    "user_is_verified",
    "user_profile_pic_id",
    "user_profile_pic_url",
    "user_fbid_v2",
]

INSTAGRAM_HASHTAG_POST_CSV_HEADERS = [
    "id",
    "media_type",
    "shortcode",
    "url",
    "caption",
    "hashtags",
    "mentioned_names",
    "liked_by_count",
    "comments_disabled",
    "comment_count",
    "preview_like_count",
    "video_view_count",
    "owner_id",
    "taken_at_timestamp",
    "utc_time",
    "accessibility_caption",
    "display_url",
]

INSTAGRAM_POST_CSV_HEADERS = [
    "username",
    "full_name",
    "is_verified",
    "id",
    "media_type",
    "shortcode",
    "url",
    "caption",
    "hashtags",
    "mentioned_names",
    "like_and_view_counts_disabled",
    "like_count",
    "comment_count",
    "main_thumbnail_url",
    "medias_type",
    "medias_url",
    "usertags_medias",
    "taken_at_timestamp",
    "utc_time",
    "coauthor_usernames",
]

INSTAGRAM_USER_CSV_HEADERS = [
    "id",
    "username",
    "full_name",
    "is_private",
    "profile_pic_url",
    "profile_pic_id",
    "is_verified",
    "has_anonymous_profile_picture",
    "has_highlight_reels",
    "account_badges",
    "similar_user_id",
    "latest_reel_media",
    "is_favorite",
]

INSTAGRAM_USER_INFO_CSV_HEADERS = [
    "username",
    "id",
    "full_name",
    "profile_pic_url_hd",
    "is_private",
    "is_verified",
    "biography",
    "bio_links_title",
    "bio_links_url",
    "biography_with_username",
    "biography_with_hashtag",
    "follower_count",
    "follow_count",
    "hide_like_and_view_counts",
    "pronouns",
    "has_ar_effects",
    "has_clips",
    "has_guides",
    "has_channel",
    "is_business_account",
    "is_professional_account",
    "is_supervision_enabled",
    "is_guardian_of_viewer",
    "is_supervised_by_viewer",
    "is_supervised_user",
    "guardian_id",
    "is_joined_recently",
    "business_category_name",
    "category_name",
    "connected_fb_page",
]



# =============================================================================
# Minet YouTube Constants
# =============================================================================
#
# General constants used throughout the YouTube functions.
#

YOUTUBE_API_BASE_URL = "https://www.googleapis.com/youtube/v3"
YOUTUBE_API_MAX_VIDEOS_PER_CALL = 50
YOUTUBE_API_MAX_CHANNELS_PER_CALL = 50
YOUTUBE_API_MAX_COMMENTS_PER_CALL = 100

YOUTUBE_API_SEARCH_ORDERS = {
    "relevance",
    "date",
    "rating",
    "viewCount",
    "title",
    "videoCount",
}

YOUTUBE_API_DEFAULT_SEARCH_ORDER = "relevance"



from typing import Optional, Iterable

from ural import URLFormatter

from minet.youtube.constants import (
    YOUTUBE_API_BASE_URL,
    YOUTUBE_API_MAX_VIDEOS_PER_CALL,
    YOUTUBE_API_MAX_COMMENTS_PER_CALL,
    YOUTUBE_API_DEFAULT_SEARCH_ORDER,
)


class YouTubeAPIURLFormatter(URLFormatter):
    BASE_URL = YOUTUBE_API_BASE_URL

    def format_arg_value(self, _, v):
        if isinstance(v, list):
            return ",".join(v)

        return v

    def playlist_videos(self, playlist_id: str, token: Optional[str]) -> str:
        return self.format(
            path="playlistItems",
            args={
                "part": "snippet",
                "playlistId": playlist_id,
                "maxResults": YOUTUBE_API_MAX_VIDEOS_PER_CALL,
                "pageToken": token,
            },
        )

    def videos(self, ids: Iterable[str]) -> str:
        return self.format(
            path="videos",
            args={"id": ids, "part": ["snippet", "statistics", "contentDetails"]},
        )

    def channels(self, ids: Iterable[str]) -> str:
        return self.format(
            path="channels",
            args={
                "id": ids,
                "part": [
                    "snippet",
                    "statistics",
                    "contentDetails",
                    "topicDetails",
                    "brandingSettings",
                    "status",
                ],
            },
        )

    def search(
        self,
        query: str,
        order: str = YOUTUBE_API_DEFAULT_SEARCH_ORDER,
        token: Optional[str] = None,
    ) -> str:
        return self.format(
            path="search",
            args={
                "part": "snippet",
                "maxResults": YOUTUBE_API_MAX_VIDEOS_PER_CALL,
                "q": query,
                "type": "video",
                "order": order,
                "pageToken": token,
            },
        )

    def comments(self, video_id: str, token: Optional[str] = None) -> str:
        return self.format(
            path="commentThreads",
            args={
                "videoId": video_id,
                "part": ["snippet", "replies"],
                "maxResults": YOUTUBE_API_MAX_COMMENTS_PER_CALL,
                "pageToken": token,
            },
        )

    def replies(self, comment_id: str, token: Optional[str] = None) -> str:
        return self.format(
            path="comments",
            args={
                "parentId": comment_id,
                "part": "snippet",
                "maxResults": YOUTUBE_API_MAX_COMMENTS_PER_CALL,
                "pageToken": token,
            },
        )



from minet.cli.argparse import command

USER_AGENTS_UPDATE_SUBCOMMAND = command(
    "update",
    "minet.cli.user_agents.update",
    title="Minet User Agents Update Command",
    description="""
        Command used to update the internal list of
        plausible User Agent strings used by minet
        to spoof some requests.

        The list of user agents come from the website:
        www.useragents.me
    """,
    no_output=True,
)

USER_AGENTS_EXPORT_SUBCOMMAND = command(
    "export",
    "minet.cli.user_agents.export",
    title="Minet User Agents Export Command",
    description="""
        Command used to export the internal list of
        plausible User Agent strings used by minet
        to spoof some requests.

        The list of user agents come from the website:
        www.useragents.me
    """,
)

USER_AGENTS_RANDOM_SUBCOMMAND = command(
    "random",
    "minet.cli.user_agents.random",
    title="Minet User Agents Random Command",
    description="""
        Command used to return a random plausible
        User Agent.

        The list of user agents come from the website:
        www.useragents.me
    """,
)

USER_AGENTS_COMMAND = command(
    "user-agents",
    "minet.cli.user_agents",
    title="Minet User Agents Command",
    aliases=["ua"],
    subcommands=[
        USER_AGENTS_UPDATE_SUBCOMMAND,
        USER_AGENTS_EXPORT_SUBCOMMAND,
        USER_AGENTS_RANDOM_SUBCOMMAND,
    ],
)



from typing import Optional, Dict, List, Union
from minet.types import AnyTimeout, RedirectionStack, Redirection, HTTPHeaderDict

import pycurl
import certifi
from dataclasses import dataclass
from ebbe import format_repr, format_filesize
from threading import Event
from ural import is_url
from urllib3 import Timeout
from urllib.parse import urljoin
from urllib3.util.url import parse_url

from minet.constants import REDIRECT_STATUSES
from minet.exceptions import (
    CancelledRequestError,
    MaxRedirectsError,
    InvalidURLError,
    PycurlError,
    PycurlHostResolutionError,
    PycurlTimeoutError,
    PycurlConnectionRefusedError,
    PycurlSSLError,
    PycurlReceiveError,
    PycurlSendError,
)

pycurl.global_init(pycurl.GLOBAL_ALL)

SHARE = pycurl.CurlShare()
SHARE.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_DNS)
SHARE.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_SSL_SESSION)
SHARE.setopt(pycurl.SH_SHARE, pycurl.LOCK_DATA_CONNECT)

CURL_HANDLE_POOL: List[pycurl.Curl] = []


@dataclass
class PycurlResult:
    url: str
    body: bytes
    headers: HTTPHeaderDict
    status: int
    stack: RedirectionStack

    def __repr__(self) -> str:
        return format_repr(
            self, ["url", "status", ("size", format_filesize(len(self.body)))]
        )


def sanitize_url(url: str) -> str:
    return parse_url(url).url


def coerce_error(
    error: pycurl.error,
) -> Union[PycurlError, CancelledRequestError, MaxRedirectsError]:
    code = error.args[0]

    if code == pycurl.E_ABORTED_BY_CALLBACK:
        return CancelledRequestError()

    if code == pycurl.E_TOO_MANY_REDIRECTS:
        return MaxRedirectsError()

    if code == pycurl.E_OPERATION_TIMEDOUT:
        return PycurlTimeoutError(error)

    if code == pycurl.E_COULDNT_RESOLVE_HOST:
        return PycurlHostResolutionError(error)

    if code == pycurl.E_COULDNT_CONNECT:
        return PycurlConnectionRefusedError(error)

    if code in (
        pycurl.E_SSL_CACERT,
        pycurl.E_PEER_FAILED_VERIFICATION,
        pycurl.E_SSL_CONNECT_ERROR,
        pycurl.E_SSL_CERTPROBLEM,
        pycurl.E_SSL_CIPHER,
    ):
        return PycurlSSLError(error)

    if code == pycurl.E_RECV_ERROR:
        return PycurlReceiveError(error)

    if code == pycurl.E_SEND_ERROR:
        return PycurlSendError(error)

    return PycurlError(error)


def setup_curl_handle(
    curl: pycurl.Curl,
    url: str,
    response_headers: HTTPHeaderDict,
    stack: RedirectionStack,
    method: str = "GET",
    headers: Optional[Dict[str, str]] = None,
    follow_redirects: bool = True,
    max_redirects: int = 5,
    compressed: bool = False,
    timeout: Optional[AnyTimeout] = None,
    cancel_event: Optional[Event] = None,
    verbose: bool = False,
) -> None:
    # Basics
    curl.setopt(pycurl.URL, sanitize_url(url))
    curl.setopt(pycurl.CAINFO, certifi.where())

    # NOTE: this is important for multithreading
    curl.setopt(pycurl.NOSIGNAL, True)

    if verbose:
        curl.setopt(pycurl.VERBOSE, True)

    if compressed:
        curl.setopt(pycurl.ACCEPT_ENCODING, "")

    # Method
    method = method.upper()

    if method == "GET":
        curl.setopt(pycurl.HTTPGET, True)
    elif method == "HEAD":
        curl.setopt(pycurl.NOBODY, True)
    elif method == "POST":
        curl.setopt(pycurl.POST, True)
    elif method == "PUT":
        curl.setopt(pycurl.PUT, True)
    else:
        curl.setopt(pycurl.CUSTOMREQUEST, method)

    # Timeout
    if timeout is not None:
        if isinstance(timeout, Timeout):
            total_timeout = None

            if timeout.total is not None:
                total_timeout = timeout.total
            else:
                if timeout.read_timeout is not None:
                    total_timeout = timeout.read_timeout
                if timeout.connect_timeout is not None:
                    if total_timeout is not None:
                        total_timeout += timeout.connect_timeout
                    else:
                        total_timeout = timeout.connect_timeout

            if total_timeout is not None:
                curl.setopt(pycurl.TIMEOUT_MS, int(total_timeout * 1000))

            if timeout.connect_timeout is not None:
                curl.setopt(
                    pycurl.CONNECTTIMEOUT_MS, int(timeout.connect_timeout * 1000)
                )

        else:
            curl.setopt(pycurl.TIMEOUT_MS, int(timeout * 1000))

    # Writing headers
    curl_headers = []

    if headers is not None:
        for n, v in headers.items():
            curl_headers.append(b"%s: %s" % (n.encode("ascii"), v.encode("latin1")))

    # NOTE: we must override Expect and Pragma to avoid unwanted libcurl magic
    if headers is None or "Expect" not in headers:
        curl_headers.append(b"Expect:")

    if headers is None or "Pragma" not in headers:
        curl_headers.append(b"Pragma:")

    curl.setopt(pycurl.HTTPHEADER, curl_headers)

    # Reading headers
    current_url = url
    expecting_location_with_status: Optional[int] = None

    def header_function(header_line):
        nonlocal current_url
        nonlocal expecting_location_with_status

        header_line = header_line.rstrip()
        header_line = header_line.decode("iso-8859-1")

        # Detecting new call, resetting headers
        if header_line.startswith("HTTP/"):
            response_headers.clear()

            status = int(header_line.split(" ", 1)[1][:3])

            if status in REDIRECT_STATUSES:
                expecting_location_with_status = status

            return

        if ":" not in header_line:
            return

        name, value = header_line.split(":", 1)

        name = name.strip()
        value = value.strip()

        if expecting_location_with_status is not None and name.lower() == "location":
            stack.append(
                Redirection(
                    current_url,
                    "location-header",
                    status=expecting_location_with_status,
                )
            )
            next_url = urljoin(current_url, value).strip()
            current_url = next_url
            expecting_location_with_status = None

        response_headers[name] = value

    curl.setopt(pycurl.HEADERFUNCTION, header_function)

    # Redirections
    if follow_redirects:
        curl.setopt(pycurl.FOLLOWLOCATION, True)
        curl.setopt(pycurl.MAXREDIRS, max_redirects)

    # Cancellation
    if cancel_event is not None:
        curl.setopt(pycurl.NOPROGRESS, False)

        def progress_function(
            download_total, downloaded_total, upload_total, uploaded_total
        ) -> int:
            if cancel_event.is_set():
                return 1

            return 0

        curl.setopt(pycurl.XFERINFOFUNCTION, progress_function)


# TODO: body
# TODO: invalid status error (pycurl has a way I think)?
# TODO: error serialization, error retrying conversion
# TODO: pool of curl handles with multi (tricks from https://github.com/tornadoweb/tornado/blob/master/tornado/curl_httpclient.py)
def request_with_pycurl(
    url: str,
    method: str = "GET",
    headers: Optional[Dict[str, str]] = None,
    follow_redirects: bool = True,
    max_redirects: int = 5,
    compressed: bool = False,
    timeout: Optional[AnyTimeout] = None,
    cancel_event: Optional[Event] = None,
    verbose: bool = False,
) -> PycurlResult:
    # Preemptive cancellation
    if cancel_event is not None and cancel_event.is_set():
        raise CancelledRequestError

    if not is_url(
        url, require_protocol=True, tld_aware=True, allow_spaces_in_path=True
    ):
        raise InvalidURLError(url)

    try:
        curl = CURL_HANDLE_POOL.pop()
        curl.reset()
    except IndexError:
        curl = pycurl.Curl()
        curl.setopt(pycurl.SHARE, SHARE)

    response_headers = HTTPHeaderDict()
    stack = []

    setup_curl_handle(
        curl,
        url=url,
        response_headers=response_headers,
        stack=stack,
        method=method,
        headers=headers,
        follow_redirects=follow_redirects,
        max_redirects=max_redirects,
        compressed=compressed,
        timeout=timeout,
        cancel_event=cancel_event,
        verbose=verbose,
    )

    # Performing
    try:
        body = curl.perform_rb()
    except pycurl.error as error:
        CURL_HANDLE_POOL.append(curl)

        raise coerce_error(error)

    status = curl.getinfo(pycurl.HTTP_CODE)
    effective_url = curl.getinfo(pycurl.EFFECTIVE_URL)
    CURL_HANDLE_POOL.append(curl)

    stack.append(Redirection(effective_url, status=status))

    return PycurlResult(
        url=url,
        body=body,
        headers=response_headers,
        status=status,
        stack=stack,
    )



# =============================================================================
# Minet Spiders
# =============================================================================
#
#
from typing import (
    TYPE_CHECKING,
    TypeVar,
    Optional,
    Tuple,
    Union,
    Callable,
    Iterable,
    Iterator,
    Generic,
    List,
)
from minet.types import ParamSpec

if TYPE_CHECKING:
    from minet.crawl.crawler import Crawler

from dataclasses import dataclass
from casanova import TabularRecord
from casanova.types import AnyWritableCSVRowPart

from minet.crawl.types import (
    UrlOrCrawlTarget,
    CrawlTarget,
    CrawlJob,
    CrawlJobDataType,
    CrawlResultDataType,
    SuccessfulCrawlResult,
)
from minet.web import Response
from minet.utils import PseudoFStringFormatter

P = ParamSpec("P")
T = TypeVar("T")

FORMATTER = PseudoFStringFormatter()

SpiderNextTargets = Optional[Iterable[UrlOrCrawlTarget[CrawlJobDataType]]]
SpiderResult = Optional[Tuple[CrawlResultDataType, SpiderNextTargets[CrawlJobDataType]]]


class Spider(Generic[CrawlJobDataType, CrawlResultDataType]):
    START_URL: Optional[str] = None
    START_URLS: Optional[Iterable[str]] = None
    START_TARGET: Optional[CrawlTarget[CrawlJobDataType]] = None
    START_TARGETS: Optional[Iterable[CrawlTarget[CrawlJobDataType]]] = None

    @property
    def crawler(self) -> "Crawler":
        c = getattr(self, "_crawler", None)

        if c is None:
            raise RuntimeError("spider has no attached crawler")

        return c

    # TODO: add some kind of method decorator?
    @property
    def is_attached(self) -> bool:
        return bool(getattr(self, "_crawler", None))

    def attach(self, crawler: "Crawler"):
        if getattr(self, "_crawler", None) is not None:
            raise TypeError("spider is already attached to a crawler")

        self._crawler = crawler

    def detach(self):
        if getattr(self, "_crawler", None) is None:
            raise TypeError("spider is not attached to a crawler")

    def start(self) -> Optional[Iterable[UrlOrCrawlTarget[CrawlJobDataType]]]:
        return None

    def process(
        self, job: CrawlJob[CrawlJobDataType], response: Response
    ) -> SpiderResult[CrawlResultDataType, CrawlJobDataType]:
        if callable(self):
            return self(job, response)

        raise NotImplementedError

    def __repr__(self):
        class_name = self.__class__.__name__

        return "<%(class_name)s>" % {"class_name": class_name}

    def write(
        self,
        filename: str,
        contents: Union[str, bytes],
        compress: bool = False,
        relative: bool = False,
    ) -> str:
        return self.crawler.write(
            filename, contents, compress=compress, relative=relative
        )

    def submit(self, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        return self.crawler.submit(fn, *args, **kwargs)

    def tabulate(
        self, result: SuccessfulCrawlResult[CrawlJobDataType, CrawlResultDataType]
    ) -> Iterator[AnyWritableCSVRowPart]:
        if result.data is None:
            return

        yield result.data  # type: ignore


FunctionSpiderCallable = Callable[
    [CrawlJob[CrawlJobDataType], Response],
    SpiderResult[CrawlResultDataType, CrawlJobDataType],
]


class FunctionSpider(Spider[CrawlJobDataType, CrawlResultDataType]):
    fn: FunctionSpiderCallable[CrawlJobDataType, CrawlResultDataType]

    def __init__(
        self,
        fn: FunctionSpiderCallable[CrawlJobDataType, CrawlResultDataType],
        start: Optional[Iterable[UrlOrCrawlTarget[CrawlJobDataType]]] = None,
    ):
        self.fn = fn
        self.__start_targets = start

    def start(self) -> Optional[Iterable[UrlOrCrawlTarget[CrawlJobDataType]]]:
        return self.__start_targets

    def process(
        self, job: CrawlJob[CrawlJobDataType], response: Response
    ) -> SpiderResult[CrawlResultDataType, CrawlJobDataType]:
        return self.fn(job, response)


# TODO: concretize output data type
@dataclass
class BasicSpiderLink(TabularRecord):
    from_url: str
    to_url: str


BasicSpiderLinks = List[BasicSpiderLink]


class BasicSpider(Spider[CrawlJobDataType, BasicSpiderLinks]):
    def process(
        self, job: CrawlJob[CrawlJobDataType], response: Response
    ) -> SpiderResult[BasicSpiderLinks, CrawlJobDataType]:
        if response.status != 200:
            return

        if not response.is_html:
            return

        next_urls = response.links(strip_fragment=True)

        links = [BasicSpiderLink(response.end_url, url) for url in next_urls]

        return links, next_urls

    def tabulate(
        self, result: SuccessfulCrawlResult[CrawlJobDataType, BasicSpiderLinks]
    ) -> Iterator[AnyWritableCSVRowPart]:
        if result.data is None:
            return

        yield from result.data



# =============================================================================
# Minet Hyphe Dump CLI Action
# =============================================================================
#
# Logic of the `hyphe dump` action.
#
import os
import csv
import base64
import zlib
import gzip
from os.path import join, dirname

from minet.utils import md5
from minet.cli.loading_bar import LoadingBar
from minet.cli.hyphe.utils import with_hyphe_fatal_errors
from minet.hyphe import HypheAPIClient
from minet.hyphe.formatters import format_webentity_for_csv, format_page_for_csv
from minet.hyphe.constants import WEBENTITY_CSV_HEADERS, PAGE_CSV_HEADERS

ADDITIONAL_PAGE_HEADERS = ["filename"]


def format_page_filename(webentity, page):
    h = md5(page["url"])

    # TODO: could be something other than html?
    return "%s/%s/%s.html.gz" % (webentity["id"], h[:2], h)


@with_hyphe_fatal_errors
def action(cli_args):
    # Paths
    output_dir = "hyphe_corpus_%s" % cli_args.corpus

    if cli_args.output_dir is not None:
        output_dir = cli_args.output_dir

    os.makedirs(output_dir, exist_ok=True)

    webentities_output_path = join(output_dir, "webentities.csv")
    pages_output_path = join(output_dir, "pages.csv")

    if cli_args.body:
        body_output_dir = join(output_dir, "content")
        os.makedirs(body_output_dir, exist_ok=True)

    client = HypheAPIClient(cli_args.url)
    corpus = client.corpus(cli_args.corpus, password=cli_args.password)

    corpus.ensure_is_started()

    # Then we gather some handy statistics
    counts = corpus.count(statuses=cli_args.statuses)

    # Then we fetch webentities
    webentities_file = open(webentities_output_path, "w", encoding="utf-8")
    webentities_writer = csv.writer(webentities_file)
    webentities_writer.writerow(WEBENTITY_CSV_HEADERS)

    with LoadingBar(
        title="Paginating web entities", unit="webentities", total=counts["webentities"]
    ) as loading_bar:
        webentities = {}

        for webentity in corpus.webentities(statuses=cli_args.statuses):
            webentities[webentity["id"]] = webentity
            webentities_writer.writerow(format_webentity_for_csv(webentity))
            loading_bar.advance()

        webentities_file.close()

    # Finally we paginate pages
    pages_file = open(pages_output_path, "w", encoding="utf-8")
    pages_writer = csv.writer(pages_file)
    pages_writer.writerow(
        PAGE_CSV_HEADERS + (ADDITIONAL_PAGE_HEADERS if cli_args.body else [])
    )

    with LoadingBar(
        title="Downloading pages", unit="pages", total=counts["pages"]
    ) as loading_bar:
        for webentity in webentities.values():
            for page in corpus.webentity_pages(
                webentity["id"], include_body=cli_args.body, count=cli_args.page_count
            ):
                filename = None

                if cli_args.body and "body" in page:
                    filename = format_page_filename(webentity, page)
                    filepath = join(body_output_dir, filename)
                    os.makedirs(dirname(filepath), exist_ok=True)

                    with open(filepath, "wb") as f:
                        binary = base64.b64decode(page["body"])
                        binary = zlib.decompress(binary)
                        binary = gzip.compress(binary)  # TODO: use gzip.open rather

                        f.write(binary)

                pages_writer.writerow(
                    format_page_for_csv(
                        webentity,
                        page,
                        expect_filename=cli_args.body,
                        filename=filename,
                    )
                )

                loading_bar.advance()



from typing import List, Optional, Dict, Tuple, Iterable

from dataclasses import dataclass
from casanova import TabularRecord
from ebbe import getpath


@dataclass
class FacebookCommentAuthor(TabularRecord):
    name: str
    short_name: str
    gender: str
    url: str
    is_verified: bool

    @classmethod
    def from_graphql_node(cls, node) -> "FacebookCommentAuthor":
        return cls(
            name=node["name"],
            short_name=node["short_name"],
            gender=node["gender"],
            url=node["url"],
            is_verified=node["is_verified"],
        )


@dataclass
class FacebookComment(TabularRecord):
    id: str
    fbid: str
    depth: int
    parent_id: Optional[str]
    author: FacebookCommentAuthor
    created_time: int
    text: str
    reactions: int
    replies: int

    @classmethod
    def from_graphql_node(cls, node) -> "FacebookComment":
        feedback = node["feedback"]

        return cls(
            id=node["id"],
            fbid=node["legacy_fbid"],
            depth=node.get("depth", 0),
            parent_id=getpath(
                node,
                ("reply_parent_comment", "id"),
                getpath(node, ("comment_parent", "id")),
            ),
            author=FacebookCommentAuthor.from_graphql_node(node["author"]),
            created_time=node["created_time"],
            text=node["preferred_body"]["text"],
            reactions=feedback["reactors"]["count"],
            replies=feedback.get("total_comment_count", 0),
        )

    @classmethod
    def from_payload(cls, payload) -> List["FacebookComment"]:
        comments = []

        # Top-level
        edges = getpath(payload, ("data", "node", "display_comments", "edges"))

        # Replies
        if edges is None:
            edges = getpath(payload, ("display_comments", "edges"))

        # Sub-replies
        if edges is None:
            edges = getpath(payload, ("data", "feedback", "display_comments", "edges"))

        # Tail call
        if not isinstance(edges, list):
            return comments

        for edge in edges:
            node = edge["node"]
            comments.append(cls.from_graphql_node(node))

            # Recursion
            feedback = node["feedback"]
            comments.extend(cls.from_payload(feedback))

        return comments

    @staticmethod
    def sort(comments: List["FacebookComment"]) -> List["FacebookComment"]:
        index: Dict[str, Tuple["FacebookComment", List["FacebookComment"]]] = {}

        for comment in sorted(comments, key=lambda c: c.depth):
            index[comment.id] = (comment, [])

            if comment.parent_id is not None:
                index[comment.parent_id][1].append(comment)

        sorted_comments: List["FacebookComment"] = []

        # NOTE: we sort in visual DFS order
        def walk(entries: Iterable[Tuple["FacebookComment", List["FacebookComment"]]]):
            for comment, replies in entries:
                sorted_comments.append(comment)

                walk((index[reply.id] for reply in replies))

        walk(filter(lambda entry: entry[0].depth == 0, index.values()))

        return sorted_comments
